#%%
from pathlib import Path
import json
from typing import Union
import shutil

class Configuration:
    def __init__(self,
                 telkey: str = None,
                 configpath: Union[Path, str] = Path.home()/'ezphot/config'):#Path(__file__).resolve().parent):
        if not Path(configpath).exists():
            print(f"[CRITICAL] Configuration path {configpath} does not exist. Run initialization first.")
        self.telkey = telkey
        self.config = dict()

        # Set up paths
        self.path_home = Path.home()
        self.path_base = Path(__file__).resolve().parent.parent
        self.path_config = Path(configpath)
        self.path_ezphot =  self.path_config.parent
        self.path_log = self.path_ezphot / 'log'
        self.path_data = self.path_ezphot / 'data'
        
        # Gloabl configuration files
        self.path_config_global = self.path_config / 'common'
        self._configfiles_global = list(self.path_config_global.glob('*.config'))
        config_global = self._load_configuration(self._configfiles_global)
        self.config.update(config_global)
        
        # Telescope specific configuration files
        self.path_config_specific = self.path_config / 'specific'
        if self.telkey:
            self.path_config_specific_telescope = self.path_config_specific / self.telkey
            self._configfiles_telescopes = list(self.path_config_specific_telescope.glob('*.config'))
            if not self._configfiles_telescopes:
                print('No configuration file is found.\nTo make default configuration files, run tcspy.configuration.make_config')
            else:
                config_unit = self._load_configuration(self._configfiles_telescopes)
                self.config.update(config_unit)
                
    def initialize(self, copy_default: bool = True):
        """Initialize the configuration by creating necessary config files."""
        # Make sure base paths exist
        self.path_config_global.mkdir(parents=True, exist_ok=True)
        self.path_config_specific.mkdir(parents=True, exist_ok=True)
        self.path_log.mkdir(parents=True, exist_ok=True)
        self.path_data.mkdir(parents=True, exist_ok=True)

        print(f"Global configuration path created: {self.path_config_global}")
        print(f"Specific configuration path created: {self.path_config_specific}")
        
        default_global_config_path = self.path_base / 'configuration' /  'common'
        default_specific_config_path = self.path_base / 'configuration' / 'specific'
        # Copy default config files to the folder
        if copy_default:
            shutil.copytree(default_global_config_path, self.path_config_global, dirs_exist_ok=True)
            print(f"Copied default global configs from {default_global_config_path}")
            shutil.copytree(default_specific_config_path, self.path_config_specific, dirs_exist_ok=True)
            print(f"Copied default telescope configs from {default_specific_config_path}")
            telescope_keys = ['\n' + p.name for p in default_specific_config_path.iterdir() if p.is_dir()]
            telescope_keys_str = ' '.join(telescope_keys)
            print(f'Current available telescope keys: {telescope_keys_str}')
            
        for tel_key in self.available_telescope_keys:
            # Rsgister each telescope
            self.telkey = tel_key
            self.path_config_specific_telescope = self.path_config_specific / self.telkey
            self.path_config_specific_telescope.mkdir(parents=True, exist_ok=True)
            self._register_telescope()
        
        self = Configuration()
        # After creating all config files, load them and make sure all directories exist
        for path in list(self.config.values()):
            self._ensure_dirs_exist(path)
            
    def register_telescope(self):
        if not self.telkey:
            raise ValueError("Telescope key (telkey) must be provided to initialize configuration.")

        self.path_config_specific_telescope.mkdir(parents=True, exist_ok=True)

        # Create config files
        self._register_telescope()
        
        for path in list(self.config.values()):
            self._ensure_dirs_exist(path)
            
    @property
    def available_telescope_keys(self):
        """Return a list of available telescope keys."""
        return [p.name for p in self.path_config_specific.iterdir() if p.is_dir()]
        
    def _load_configuration(self, configfiles):
        all_config = dict()
        for configfile in configfiles:
            with open(configfile, 'r') as f:
                config = json.load(f)
                all_config.update(config)
        return all_config

    def _make_configfile(self, dict_params: dict, filename: str, savepath: Union[str, Path]):
        filepath = Path(savepath) / filename
        with open(filepath, 'w') as f:
            json.dump(dict_params, f, indent=4)
        print(f'New configuration file made: {filepath}')

    def _ensure_dirs_exist(self, *paths):
        """Create all directories in the given paths if they don't exist."""
        for p in paths:
            try:
                Path(p).mkdir(parents=True, exist_ok=True)
            except Exception as e:
                pass
                #print(f"[WARNING] Could not create directory {p}: {e}")
                
    def _register_telescope(self):
        
        # Specific telescope configuration
        sex_config = dict(
            SEX_CONFIG = str(self.path_config_global / 'sextractor' / f'{self.telkey}.sexconfig'),
            SEX_CONFIGDIR = str(self.path_config_global / 'sextractor'),
            SEX_LOGDIR = str(self.path_log / 'sextractor' / 'log'),
            SEX_HISTORYDIR = str(self.path_log / 'sextractor' / 'history')
        )
        astrometry_config = dict(
            ASTROMETRY_SEXCONFIG = str(self.path_config_global / 'sextractor' / f'{self.telkey}.astrometry.sexconfig')
        )
        scamp_config = dict(
            SCAMP_CONFIG = str(self.path_config_global / 'scamp' / 'default.scampconfig'),
            SCAMP_SEXCONFIG = str(self.path_config_global / 'sextractor' / f'{self.telkey}.scamp.sexconfig'),
            SCAMP_CONFIGDIR = str(self.path_config_global / 'scamp'),
            SCAMP_LOGDIR = str(self.path_log / 'scamp' / 'log'),
            SCAMP_HISTORYDIR = str(self.path_log / 'scamp' / 'history')
        )
        swarp_config = dict(
            SWARP_CONFIG = str(self.path_config_global / 'swarp' / f'{self.telkey}.swarpconfig'),
            SWARP_CONFIGDIR = str(self.path_config_global / 'swarp'),
            SWARP_LOGDIR = str(self.path_log / 'swarp' / 'log'),
            SWARP_HISTORYDIR = str(self.path_log / 'swarp' / 'history')
        )
        psfex_config = dict(
            PSFEX_CONFIG = str(self.path_config_global / 'psfex' / 'default.psfexconfig'),
            PSFEX_SEXCONFIG = str(self.path_config_global / 'sextractor' / f'{self.telkey}.psfex.sexconfig'),
            PSFEX_CONFIGDIR = str(self.path_config_global / 'psfex'),
            PSFEX_LOGDIR = str(self.path_log / 'psfex' / 'log'),
            PSFEX_HISTORYDIR = str(self.path_log / 'psfex' / 'history')
        )
        
        for cfg, name in [
            (sex_config, 'sex.config'),
            (astrometry_config, 'astrometry.config'),
            (scamp_config, 'scamp.config'),
            (swarp_config, 'swarp.config'),
            (psfex_config, 'psfex.config')
        ]:
            self._make_configfile(cfg, name, self.path_config_specific_telescope)

        # Global configuration
        calibdata_config = dict(
            CALIBDATA_DIR = str(self.path_data / 'calibdata'),
            CALIBDATA_MASTERDIR = str(self.path_data / 'mcalibdata'),
            )
        refdata_config = dict(
            REFDATA_DIR = str(self.path_data / 'refdata'),
            )
        obsdata_config = dict(OBSDATA_DIR = str(self.path_data / 'obsdata'))
        scidata_config = dict(SCIDATA_DIR = str(self.path_data / 'scidata'))
        catalog_config = dict(CATALOG_DIR = str(self.path_data / 'skycatalog' / 'archive'))
        observatory_config = dict(
            OBSERVATORY_LOCATIONINFO = str(self.path_config_global / 'obs_location.txt'),
            OBSERVATORY_TELESCOPEINFO = str(self.path_config_global / 'CCD.dat')
        )
        sdtdata_config = dict(
            SDTDATA_OBSSOURCEDIR = str(self.path_data / 'connecteddata' / '7DT' / 'obsdata'),
            SDTDATA_OBSDESTDIR = str(self.path_data / 'obsdata' / '7DT'),
            SDTDATA_SCISOURCEDIR = str(self.path_data / 'connecteddata' / '7DT' / 'processed_1x1_gain2750'),
            SDTDATA_SCIDESTDIR = str(self.path_data / 'scidata' / '7DT' / '7DT_C361K_HIGH_1x1')
        )

        for cfg, name in [
            (calibdata_config, 'calibdata.config'),
            (refdata_config, 'refdata.config'),
            (scidata_config, 'scidata.config'),
            (catalog_config, 'catalog.config'),
            (observatory_config, 'observatory.config'),
            (sdtdata_config, 'sdtdata.config'),
            (obsdata_config, 'obsdata.config')
        ]:
            self._make_configfile(cfg, name, self.path_config_global)

        # Remove per-telescope specific keys before saving global versions
        sex_config.pop('SEX_CONFIG', None)
        scamp_config.pop('SCAMP_SEXCONFIG', None)
        swarp_config.pop('SWARP_CONFIG', None)
        psfex_config.pop('PSFEX_SEXCONFIG', None)

        for cfg, name in [
            (sex_config, 'sex.config'),
            (scamp_config, 'scamp.config'),
            (swarp_config, 'swarp.config'),
            (psfex_config, 'psfex.config'),
        ]:
            self._make_configfile(cfg, name, self.path_config_global)
#%%
if __name__ == '__main__':
    # For the first time, initiailize the configuration
    config = Configuration()
    config.initialize(copy_default=True)
    
    telescope_keys = [
        '7DT_C361K_HIGH_1x1', '7DT_C361K_HIGH_2x2', '7DT_C361K_LOW_1x1', '7DT_C361K_LOW_2x2',
        'CBNUO_STX16803_1x1', 'LSGT_SNUCAMII_1x1', 'LSGT_ASI1600MM_1x1',
        'RASA36_KL4040_HIGH_1x1', 'RASA36_KL4040_MERGE_1x1', 'SAO_C361K_1x1',
        'SOAO_FLI4K_1x1', 'KCT_STX16803_1x1', 'SkyMapper_SG_32_Det_1x1']
    
    for key in telescope_keys:
        print(key)
        config = Configuration(telkey=key)
        config.register_telescope()
        #print(config.config)

# %%



#%%
import inspect
from typing import Union, Optional
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, Ellipse, Circle
from tqdm import tqdm
from scipy.ndimage import mean as ndi_mean
from astropy.table import Table
from astropy.modeling import models, fitting
from astropy.coordinates import SkyCoord
from astropy.wcs.utils import pixel_to_skycoord, skycoord_to_pixel
from photutils.segmentation import detect_sources, deblend_sources, SourceCatalog
from photutils.aperture import (
    CircularAperture, CircularAnnulus, aperture_photometry,
    EllipticalAperture, EllipticalAnnulus
)
from photutils.utils import calc_total_error

from ezphot.methods import BackgroundGenerator, ErrormapGenerator
from ezphot.helper import Helper
from ezphot.imageobjects import (
    ScienceImage, ReferenceImage, CalibrationImage,
    Background, Mask, Errormap
)
from ezphot.dataobjects import Catalog
from ezphot.utils import *


class AperturePhotometry:
    """
    Method class for performing aperture photometry.
    
    This class provides methods 
    
    1. SExtractor-based aperture photometry.
    
    2. Photutils-based aperture photometry.
    
    3. Circular aperture photometry.
    
    4. Elliptical aperture photometry.
    
    5. Kron aperture photometry.
    
    6. Circular aperture photometry.
    
    """
    def __init__(self):
        """
        Initialize the AperturePhotometry class.
        
        Parameters
        ----------
        None
        """ 
        self.helper = Helper()
        self.background = BackgroundGenerator()
        self.errormap = ErrormapGenerator() 
        
    def __repr__(self):
        return f"Method class: {self.__class__.__name__}\n For help, use 'help(self)' or `self.help()`."

    def help(self):
        # Get all public methods from the class, excluding `help`
        methods = [
            (name, obj)
            for name, obj in inspect.getmembers(self.__class__, inspect.isfunction)
            if not name.startswith("_") and name != "help"
        ]

        # Build plain text list with parameters
        lines = []
        for name, func in methods:
            sig = inspect.signature(func)
            params = [str(p) for p in sig.parameters.values() if p.name != "self"]
            sig_str = f"({', '.join(params)})" if params else "()"
            lines.append(f"- {name}{sig_str}")

        # Final plain text output
        help_text = ""
        print(f"Help for {self.__class__.__name__}\n{help_text}\nPublic methods:\n" + "\n".join(lines))
        
    def run_photometry(self,
                       target_img: Union[ScienceImage, ReferenceImage, CalibrationImage],
                       target_bkg: Optional[Background] = None,
                       target_bkgrms: Optional[Errormap] = None,
                       target_mask: Optional[Mask] = None,
                       sex_params: dict = None,
                       detection_sigma: float = 5,
                       aperture_diameter_arcsec: Union[float, list] = [5, 7, 10],
                       aperture_diameter_seeing: Union[float, list] = [3.5, 4.5],
                       saturation_level: float = 60000,
                       kron_factor: float = 2.5,
                       save: bool = True,
                       verbose: bool = True,
                       visualize: bool = True,
                       save_fig: bool = False,
                       **kwargs
                       ):
        """
        Run photometry on the target image.
        """
        pass

    def sex_photometry(self,
                       # Input parameters
                       target_img: Union[ScienceImage, ReferenceImage, CalibrationImage], 
                       target_bkg: Optional[Background] = None, # If target_bkg is given, subtract background before sextractor
                       target_bkgrms: Optional[Errormap] = None, # It must be background error map
                       target_mask: Optional[Mask] = None, # For masking certain source (such as hot pixels)
                       sex_params: dict = None,
                       detection_sigma: float = 5,
                       aperture_diameter_arcsec: Union[float, list] = [5, 7, 10],
                       aperture_diameter_seeing: Union[float, list] = [3.5, 4.5], # If given, use seeing to calculate aperture size
                       saturation_level: float = 60000,
                       kron_factor: float = 2.5,
                       
                       # Others
                       save: bool = True,
                       verbose: bool = True,
                       visualize: bool = True,
                       save_fig: bool = False,
                       **kwargs
                       ):
        """
        Perform aperture photometry using SExtractor.

        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage, CalibrationImage]
            The target image to perform aperture photometry on.
        target_bkg : Optional[Background], optional
            The background to subtract from the target image. Defaults to None.
        target_bkgrms : Optional[Errormap], optional
            The background RMS to use for the aperture photometry. Defaults to None.
        detection_sigma : float, optional
            The detection sigma for the aperture photometry. Defaults to 5.
        aperture_diameter_arcsec : Union[float, list], optional
            The aperture diameter in arcseconds. Defaults to [5, 7, 10].
        aperture_diameter_seeing : Union[float, list], optional
            The aperture diameter in seeing units. Defaults to [3.5, 4.5].
        saturation_level : float, optional
            The saturation level for the aperture photometry. Defaults to 60000.
        kron_factor : float, optional
            The Kron factor for the aperture photometry. Defaults to 2.5.
        save : bool, optional
            Whether to save the aperture photometry catalog. Defaults to True.
        verbose : bool, optional
            Whether to print verbose output. Defaults to True.
        visualize : bool, optional
            Whether to visualize the aperture photometry. Defaults to True.
        save_fig : bool, optional
            Whether to save the aperture photometry figure. Defaults to False.

        Raises
        ------
        ValueError: If target_img is not a ScienceImage, ReferenceImage, or CalibrationImage.
        RuntimeError: If the source extractor fails.
        RuntimeError: If the source extractor fails.

        Returns
        -------
        target_catalog : Catalog
            The aperture photometry catalog.
        """
        if not isinstance(target_img, (ScienceImage, ReferenceImage, CalibrationImage)):
            raise ValueError('target_img must be a ScienceImage, ReferenceImage, or CalibrationImage.')
        
        if not target_img.is_saved:
            target_img.write(verbose = False)
        
        # If sex_params is None, use default parameters
        if sex_params is None:
            sex_params = dict()
        
        #target_img = target_img.copy()
        img_path = target_img.savepath.savepath
        cat_path = target_img.savepath.catalogpath
        sexconfig_path = target_img.config['SEX_CONFIG']
        all_sexconfig = self.helper.load_config(sexconfig_path)
        
        # If target_bkg is given, subtract background before sextractor
        remove_subbkg = False
        if target_bkg is not None:
            target_img_sub = self.background.subtract_background(
                target_img = target_img, 
                target_bkg = target_bkg,
                save = True,
                overwrite = False,
                visualize = visualize,
                save_fig = save_fig,
                verbose = verbose
            )
            saturation_level -= target_bkg.info.BKGVALU
            img_path = target_img_sub.savepath.savepath
            sex_params['BACK_TYPE'] = 'MANUAL'
            remove_subbkg = True
        else:
            target_img_sub = target_img
            
        # Background RMS
        remove_bkgrms = False
        if target_bkgrms is not None:
            if not target_bkgrms.is_saved:
                target_bkgrms.write(verbose = False)
                remove_bkgrms = True
            bkgrms_path = target_bkgrms.savepath.savepath
        else:
            bkgrms_path = None
        
        # Mask
        remove_mask = False
        if target_mask is not None:
            if not target_mask.is_saved:
                target_mask.write(verbose = False)
                remove_mask = True
            mask_path = target_mask.savepath.savepath
        else:
            mask_path = None
                    
        # First Sextractor run to estimate seeing if not provided
        if target_img.seeing is None:
            sex_params['DETECT_THRESH'] = 10
            result, catalog_first, global_bkgval, global_bkgrms = self.helper.run_sextractor(
                target_path = img_path,
                sex_configfile = sexconfig_path,
                sex_params = sex_params,
                target_mask = mask_path,
                target_weight = None,
                weight_type = 'MAP_RMS',
                target_outpath = cat_path,
                return_result = True,
                verbose = verbose
            ) 
            if not result:
                raise RuntimeError('Source extractor is failed.')
            
            seeing_estimate = None
            if 'FLAGS' in catalog_first.colnames:
                rough_flags = (catalog_first['FLAGS'] == 0)
            if 'MAG_AUTO' in catalog_first.colnames:
                rough_flags &= (catalog_first['MAG_AUTO'] < 0)
                rough_flags &= (catalog_first['MAG_AUTO'] > -20)
            if 'FLUX_MAX' in catalog_first.colnames:
                rough_flags &= (catalog_first['FLUX_MAX'] < 60000)
            if 'ELONGATION' in catalog_first.colnames:
                rough_flags &= (catalog_first['ELONGATION'] < 1.3)
            if 'CLASS_STAR' in catalog_first.colnames:
                rough_flags &= (catalog_first['CLASS_STAR'] > 0.2)
            seeing_arcsec = catalog_first['FWHM_WORLD'] * 3600
            rough_flags &= (seeing_arcsec < 10)
            rough_flags &= (seeing_arcsec > 0.5)
            catalog_filtered = catalog_first[rough_flags]
            seeing_estimate = max(1.2, np.mean(catalog_filtered['FWHM_WORLD']) * 3600)
        else:
            seeing_estimate = target_img.seeing
        
        if "SEEING_FWHM" not in sex_params.keys():
            sex_params['SEEING_FWHM'] = '%.2f' %seeing_estimate            
        sex_params['PIXEL_SCALE'] = np.mean(target_img.pixelscale)
        
        all_apertures = []
        aperture_diameter_arcsec = np.atleast_1d(aperture_diameter_arcsec)
        for aperture_size in aperture_diameter_arcsec:
            all_apertures.append(aperture_size)
        if aperture_diameter_seeing is not None:
            aperture_diameter_seeing = np.atleast_1d(aperture_diameter_seeing)
            for aperture_seeing_ratio in aperture_diameter_seeing:
                all_apertures.append(seeing_estimate * aperture_seeing_ratio)
        
        aperture_diameter_pixel = ','.join(["%.2f"%(float(size / target_img.telinfo['pixelscale'])) for size in all_apertures])
        sex_params['PHOT_APERTURES'] = aperture_diameter_pixel # This is aperture size in pixel
        sex_params['SATUR_LEVEL'] = saturation_level
        sex_params['PHOT_AUTOPARAMS'] = f"{kron_factor},3.5"
        if 'DETECT_MINAREA' in sex_params.keys():
            if sex_params['DETECT_MINAREA'] < 3:
                sex_params['DETECT_MINAREA'] = 3
                self.helper.print(f"[WARNING] DETECT_MINAREA is less than 3. It is set to 3.", verbose)
        if 'DETECT_THRESH' in sex_params.keys():
            if sex_params['DETECT_THRESH'] < 1:
                sex_params['DETECT_THRESH'] = 1
                self.helper.print('[WARNING] DETECT_THRESH is less than 1. It is set to 1.', verbose)
        for key, value in sex_params.items():
            all_sexconfig[key] = value

        # If DETECT_MINAREA is not set, use the default value from the config
        # if 'DETECT_MINAREA' not in sex_params.keys():
        #     sex_params['DETECT_MINAREA'] = all_sexconfig['DETECT_MINAREA']
        # If DETECT_THRESH is not set, use the detection_sigma
        #if 'DETECT_THRESH' not in sex_params.keys():
        #    sex_params['DETECT_THRESH'] = detection_sigma / np.sqrt(sex_params['DETECT_MINAREA'])
        #if 'ANALYSIS_THRESH' not in sex_params.keys():
        #    sex_params['ANALYSIS_THRESH'] = sex_params['DETECT_THRESH']

        
        # Second Sextractor run with the estimated parameters
        result, catalog, global_bkgval, global_bkgrms = self.helper.run_sextractor(
            target_path = img_path,
            sex_configfile = sexconfig_path,
            sex_params = sex_params.copy(),
            target_mask = mask_path,
            target_weight = bkgrms_path,
            weight_type = 'MAP_RMS',
            target_outpath = cat_path,
            return_result = True,
            verbose = verbose
        ) 
        
        if not result:
            raise RuntimeError('Source extractor is failed.')
        
        # Modification of the catalog
        catalog['X_IMAGE'] = catalog['X_IMAGE'] -1 # 0-based index
        catalog['Y_IMAGE'] = catalog['Y_IMAGE'] -1 # 0-based index
        catalog['SKYSIG'] = global_bkgrms
        catalog['SKYVAL'] = global_bkgval
        catalog['DETECT_THRESH'] = detection_sigma

        # Kron aperture area
        a = catalog['KRON_RADIUS'] * catalog['A_IMAGE']
        b = catalog['KRON_RADIUS'] * catalog['B_IMAGE']
        catalog['NPIX_AUTO'] = np.pi * a * b

        # Circular aperture areas

        pixelscale = np.mean(target_img.pixelscale)
        for i, ap_size_arcsec in enumerate(all_apertures):
            radius_pixel = ap_size_arcsec / pixelscale / 2
            area_pixel = np.pi * (radius_pixel)**2
            npix_colname = 'NPIX_APER' if i == 0 else f'NPIX_APER_{i}'
            catalog[npix_colname] = np.full(len(catalog), area_pixel)
            pixsize_colname = 'PIXSIZE_APER' if i == 0 else f'PIXSIZE_APER_{i}'
            catalog[pixsize_colname] = np.full(len(catalog), ap_size_arcsec / pixelscale)

        target_catalog = Catalog(path = cat_path, catalog_type = 'all', load = False) 
        target_catalog.data = catalog
        
        if save:
            target_catalog.write(verbose = verbose)
        else:
            target_catalog.remove(verbose = verbose)
        
        if visualize:
            save_path = None
            if save_fig:
                save_path = str(target_catalog.path) + '.png'
            self._visualize_objects(target_img=target_img_sub, 
                                    objects=target_catalog.data, 
                                    size=1000, 
                                    save_path = save_path,
                                    show = visualize)
            
        for remove_trigger, remove_object in zip(
            [remove_subbkg, remove_bkgrms, remove_mask],
            [target_img_sub, target_bkgrms, target_mask]):
            if remove_trigger:
                remove_object.remove(verbose = verbose)
        
        return target_catalog
    
    def photutils_photometry(self,
                             # Input parameters
                             target_img: Union[ScienceImage, ReferenceImage, CalibrationImage], 
                             target_bkg: Optional[Background] = None,
                             target_bkgrms: Optional[Errormap] = None,
                             target_mask: Optional[Mask] = None,
                             detection_sigma: float = 1.5,
                             aperture_diameter_arcsec: Union[float, list] = [5,7,10],
                             kron_factor: float = 2.5,
                             minarea_pixels: int = 5,
                             calc_accurate_fwhm: bool = True,
                            
                             # Other options
                             save: bool = True,
                             verbose: bool = True,
                             visualize: bool = True,
                             save_fig: bool = False,
                             **kwargs):
        """
        Perform circular and Kron photometry using photutils.SourceCatalog.
        
        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage, CalibrationImage]
            The target image to perform aperture photometry on.
        target_bkg : Optional[Background], optional
            The background to subtract from the target image. Defaults to None.
        target_bkgrms : Optional[Errormap], optional
            The background RMS to use for the aperture photometry. Defaults to None.
        target_mask : Optional[Mask], optional
            The mask to use for the aperture photometry. Defaults to None.
        detection_sigma : float, optional
            The detection sigma for the aperture photometry. Defaults to 1.5.
        aperture_diameter_arcsec : Union[float, list], optional
            The aperture diameter in arcseconds. Defaults to [5,7,10].
        kron_factor : float, optional
            The Kron factor for the aperture photometry. Defaults to 2.5.
        minarea_pixels : int, optional
            The minimum area in pixels for the aperture photometry. Defaults to 5.
        calc_accurate_fwhm : bool, optional
            Whether to calculate the accurate FWHM. Defaults to True.
        save : bool, optional
            Whether to save the aperture photometry catalog. Defaults to True.
        verbose : bool, optional
            Whether to print verbose output. Defaults to True.
        visualize : bool, optional
            Whether to visualize the aperture photometry. Defaults to True.
        save_fig : bool, optional
            Whether to save the aperture photometry figure. Defaults to False.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        target_catalog : Catalog
            The aperture photometry catalog.
        """
        if not isinstance(target_img, (ScienceImage, ReferenceImage, CalibrationImage)):
            raise ValueError("target_img must be a ScienceImage, ReferenceImage, or CalibrationImage.")
        
        bkgrms_map = None
        bkgrms = None
        mask_map = target_mask.data if target_mask is not None else None
        mask_map = self.helper.to_native(mask_map)
        target_data = self.helper.to_native(target_img.data)
        # Step 1: Background subtraction
        if target_bkg is not None:
            target_img_sub = self.background.subtract_background(
                target_img=target_img,
                target_bkg=target_bkg,
                save = False,
                overwrite=False,
                visualize = visualize,
                save_fig = save_fig,
            )
            target_data = self.helper.to_native(target_img_sub.data)
            target_bkg_data = target_bkg.data
        else:
            target_bkg_data = None

        # Step 2: Set error map
        if target_bkgrms is not None:
            bkgrms = self.helper.to_native(target_bkgrms.data)
        elif bkgrms_map is not None:
            bkgrms = self.helper.to_native(bkgrms_map.data)
        else:
            # Use sigma-clipped std if no error map available
            bkgrms_map, _, _ = self.errormap.calculate_errormap_from_image(
                target_img=target_img,
                target_mask=None,
                mode='photutils',
                errormap_type='bkgrms',
                n_iterations = 0,
                save=False,
                visualize=visualize,
                save_fig = False,
                **kwargs
            )
            bkgrms = bkgrms_map.data
        
        error = calc_total_error(data=target_data, bkg_error=bkgrms, effective_gain=target_img.egain)

        # 3. Segmentation and deblending
        threshold = detection_sigma * bkgrms
        segm = detect_sources(target_data, threshold, npixels=minarea_pixels, mask=mask_map)
        if segm is None:
            return None
        segm = deblend_sources(target_data, segm, npixels=minarea_pixels, nlevels=32, contrast=0.005)
        
        # 4. SourceCatalog (deblended)
        cat = SourceCatalog(data=target_data, segment_img=segm, error=error, mask=mask_map, background = target_bkg_data, wcs = target_img.wcs, kron_params = (kron_factor, 1.0, 0))
        cat_tbl = cat.to_table()
        cat_tbl['kron_radius'] = cat.kron_radius * kron_factor
        cat_tbl['flux_radius'] = cat.fluxfrac_radius(0.5)
        cat_tbl['ellipticity'] = cat.ellipticity
        cat_tbl['fwhm_pixel'] = 2.3548 / 1.1774 * cat_tbl['flux_radius'] # From flux radius, gaussian approximation
        coords = cat_tbl['sky_centroid']
        cat_tbl['ra'] = coords.ra.value
        cat_tbl['dec'] = coords.dec.value

        # FWHM calculation
        if calc_accurate_fwhm:
            all_fwhm = []
            for source in tqdm(cat_tbl, desc = 'Calculating FWHM...'):
                x0, y0 = source['xcentroid'], source['ycentroid']
                stamp = self.helper.img_extract_stamp(target_data, x0, y0, size=25)  # make your own stamp function
                y, x = np.indices(stamp.shape)
                
                model = models.Gaussian2D(amplitude=np.max(stamp), x_mean=12.5, y_mean=12.5, x_stddev=2, y_stddev=2)
                fitter = fitting.LevMarLSQFitter()
                fit_model = fitter(model, x, y, stamp)
                
                fwhm_x = 2.3548 * fit_model.x_stddev.value
                fwhm_y = 2.3548 * fit_model.y_stddev.value
                fwhm_avg = np.sqrt(fwhm_x * fwhm_y)
                all_fwhm.append(fwhm_avg)
            cat_tbl['fwhm_pixel'] = all_fwhm
        
        rename_map = {
            'label': 'NUMBER',
            'xcentroid': 'X_IMAGE',
            'ycentroid': 'Y_IMAGE',
            'ra': 'X_WORLD',
            'dec': 'Y_WORLD',
            'bbox_xmin': 'XMIN_IMAGE',
            'bbox_xmax': 'XMAX_IMAGE',
            'bbox_ymin': 'YMIN_IMAGE',
            'bbox_ymax': 'YMAX_IMAGE',
            'area': 'ISOAREA_IMAGE',
            'semimajor_sigma': 'A_IMAGE',
            'semiminor_sigma': 'B_IMAGE',
            'orientation': 'THETA_IMAGE', 
            'eccentricity': 'ECCENTRICITY',
            'ellipticity': 'ELLIPTRICITY',
            'min_value': 'FLUX_MIN',
            'max_value': 'FLUX_MAX',
            'segment_flux': 'FLUX_ISO',
            'segment_fluxerr': 'FLUXERR_ISO',
            'kron_flux': 'FLUX_AUTO',
            'kron_fluxerr': 'FLUXERR_AUTO',
            'kron_radius': 'KRON_RADIUS',
            'flux_radius': 'FLUX_RADIUS',
            'fwhm_pixel': 'FWHM_IMAGE',
            'max_value': 'FLUX_MAX',
            'fwhm_pixel': 'FWHM_IMAGE'
        }
        
        catalog = Table()#objects.copy()

        # Modification of the catalog
        for old, new in rename_map.items():
            if old in cat_tbl.colnames:
                catalog[new] = cat_tbl[old]
                
        catalog['MAG_AUTO'] = -2.5 * np.log10(catalog['FLUX_AUTO'])
        catalog['MAGERR_AUTO'] = 2.5 / np.log(10) * catalog['FLUXERR_AUTO'] / catalog['FLUX_AUTO']
        a = catalog['KRON_RADIUS'] * catalog['A_IMAGE']
        b = catalog['KRON_RADIUS'] * catalog['B_IMAGE']
        catalog['NPIX_AUTO'] = np.pi * a * b
        catalog['ELONGATION'] = catalog['A_IMAGE'] / catalog['B_IMAGE']
        catalog['SKYSIG'] = ndi_mean(bkgrms, labels=segm.data, index=segm.labels)
        catalog['THRESHOLD'] = catalog['SKYSIG'] * detection_sigma
        catalog['DETECT_THRESH'] = detection_sigma

        # Circular photometry 
        pixelscale = np.mean(target_img.pixelscale)
        aperture_diameter_arcsec = np.atleast_1d(aperture_diameter_arcsec)
        aperture_diameter_pixel = aperture_diameter_arcsec / pixelscale

        for i, diameter_pixel in enumerate(aperture_diameter_pixel):
            radius_pixel = diameter_pixel / 2
            circular_phot = cat.circular_photometry(radius=radius_pixel)
            suffix_key = '' if i == 0 else '_%d' % i
            catalog[f'FLUX_APER{suffix_key}'] = circular_phot[0]
            catalog[f'FLUXERR_APER{suffix_key}'] = circular_phot[1]
            catalog[f'MAG_APER{suffix_key}'] = -2.5*np.log10(circular_phot[0])
            catalog[f'MAGERR_APER{suffix_key}'] = 2.5/np.log(10) * circular_phot[1] / circular_phot[0]
            
            area_pixel = np.pi * (radius_pixel)**2
            catalog[f'NPIX_APER{suffix_key}'] = np.full(len(catalog), area_pixel)

        cat_path = target_img.savepath.catalogpath  
        target_catalog = Catalog(path = cat_path, catalog_type = 'all', load = False) 
        target_catalog.data = catalog
        
        if save:
            target_catalog.write(verbose = verbose)
        else:
            target_catalog.remove(verbose = verbose)
            
        if visualize:
            save_path = None
            if save_fig:
                save_path = str(target_catalog.path) + '.png'
            self._visualize_objects(target_img=target_img, 
                                    objects=target_catalog.data, 
                                    size=1000, 
                                    save_path = save_path,
                                    show = visualize)
            
        if target_bkg is not None:
            target_img_sub.remove(verbose = verbose)
        return target_catalog
        
    def circular_photometry(self,
                            target_img: Union[ScienceImage, ReferenceImage],
                            x_arr: Union[float, list, np.ndarray],
                            y_arr: Union[float, list, np.ndarray],
                            aperture_diameter_arcsec: Union[float, list] = [5,7,10],
                            aperture_diameter_seeing: Union[float, list] = [3.5, 4.5], # If given, use seeing to calculate aperture size
                            annulus_width_arcsec: Union[float, list] = None, # When local background is used
                            unit: str = 'pixel',
                            target_bkg: Optional[Background] = None,
                            target_mask: Union[str, Path, np.ndarray] = None,
                            target_bkgrms: Optional[Errormap] = None,
                            
                            # Other paramters
                            save: bool = True,
                            verbose: bool = True,
                            visualize: bool = True,
                            save_fig: bool = False,
                            **kwargs
                            ):
        """
        Perform circular aperture photometry.
        
        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage, CalibrationImage]
            The target image to perform aperture photometry on.
        x_arr : Union[float, list, np.ndarray]
            The x-coordinates of the sources.
        y_arr : Union[float, list, np.ndarray]
            The y-coordinates of the sources.
        aperture_diameter_arcsec : Union[float, list], optional
            The aperture diameter in arcseconds. Defaults to [5,7,10].
        annulus_diameter_arcsec : Union[float, list], optional
            The annulus diameter in arcseconds. Defaults to None.
        unit : str, optional
            The unit of the coordinates. Defaults to 'pixel'.
        target_bkg : Optional[Background], optional
            The background to subtract from the target image. Defaults to None.
        target_mask : Optional[Mask], optional
            The mask to use for the aperture photometry. Defaults to None.
        target_bkgrms : Optional[Errormap], optional
            The background RMS to use for the aperture photometry. Defaults to None.
        save : bool, optional
            Whether to save the aperture photometry catalog. Defaults to True.
        visualize : bool, optional
            Whether to visualize the aperture photometry. Defaults to True.
        save_fig : bool, optional
            Whether to save the aperture photometry figure. Defaults to False.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        target_catalog : Catalog
            The aperture photometry catalog.
        """

        # Step 1: Background subtraction
        data = self.helper.to_native(target_img.data)
        if target_bkg is not None:
            target_img_sub = self.background.subtract_background(
                target_img=target_img,
                target_bkg=target_bkg,
                save = False,
                overwrite=False,
                visualize = visualize,
                save_fig = save_fig,
            )
            data = self.helper.to_native(target_img_sub.data)
            
        # If target_bkgrms is not given, calculate it from the target image
        if target_bkgrms is None:
            target_bkgrms, _, _ = self.errormap.calculate_errormap_from_image(
                target_img = target_img,
                errormap_type = 'bkgrms',
                save = False,
                verbose = True,
                visualize = visualize,
                save_fig = False
            )
            
        # Step 2: Prepare data
        mask = self.helper.to_native(target_mask.data) if target_mask is not None else None
        bkgrms = self.helper.to_native(target_bkgrms.data)
        error = calc_total_error(data=data, bkg_error=bkgrms, effective_gain=target_img.egain)

        # Step 3: Pixel scale (arcsec/pix)
        pixelscale = np.mean(target_img.pixelscale)

        # Step 4: Normalize radius inputs
        all_apertures = []
        aperture_diameter_arcsec = np.atleast_1d(aperture_diameter_arcsec)
        for aperture_size in aperture_diameter_arcsec:
            all_apertures.append(aperture_size)
        if aperture_diameter_seeing is not None:
            if target_img.seeing is not None:
                aperture_diameter_seeing = np.atleast_1d(aperture_diameter_seeing)
                for aperture_seeing_ratio in aperture_diameter_seeing:
                    all_apertures.append(target_img.seeing * aperture_seeing_ratio)
            else:
                print("[WARNING] target_img.seeing is not defined. Using aperture_diameter_seeing is ignored.")
        aperture_diameter_pixel = all_apertures / pixelscale
        
        all_annulus = []
        if annulus_width_arcsec is not None:
            for aperture_diameter in all_apertures:
                annulus_diameter = aperture_diameter + annulus_width_arcsec
                all_annulus.append(annulus_diameter)
            print('ANNULUS applied')
        annulus_diameter_pixel = all_annulus / pixelscale if all_annulus else None

        # Step 5: Source positions
        x_arr = np.atleast_1d(x_arr)
        y_arr = np.atleast_1d(y_arr)

        skycoord = None
        if unit == 'pixel':
            positions = np.transpose((x_arr, y_arr))
            wcs = target_img.wcs
            if wcs:
                skycoord = pixel_to_skycoord(x_arr, y_arr, wcs)
        elif unit == 'coord':
            skycoord = SkyCoord(ra=x_arr, dec=y_arr, unit='deg')
            x_pix, y_pix = skycoord_to_pixel(skycoord, target_img.wcs)
            positions = np.transpose((x_pix, y_pix))
        else:
            raise ValueError("unit must be either 'pixel' or 'coord'")

        # Step 6: Photometry
        results = Table()
        results['X_IMAGE'] = positions[:, 0]
        results['Y_IMAGE'] = positions[:, 1]
        if skycoord:
            results['X_WORLD'] = skycoord.ra.value
            results['Y_WORLD'] = skycoord.dec.value
        
        for i, diameter_pixel in enumerate(aperture_diameter_pixel):
            radius_pixel = diameter_pixel /2
            aperture = CircularAperture(positions, r=radius_pixel)

            # Photometry on background-subtracted image
            phot_table = aperture_photometry(data, aperture, error=error, mask=mask)
                
            suffix_key = '' if i == 0 else '_%d'%i
            flux_key = f'FLUX_APER{suffix_key}'
            fluxerr_key = f'FLUXERR_APER{suffix_key}'
            mag_key = f'MAG_APER{suffix_key}'
            magerr_key = f'MAGERR_APER{suffix_key}'
            annul_key = f'FLUX_ANNULUS{suffix_key}'
            magannul_key = f'MAG_ANNULUS{suffix_key}'
            npix_key = f'NPIX_APER{suffix_key}'
            skysig_key = f'SKYSIG_APER{suffix_key}'
            ul5_key = f'UL5_APER{suffix_key}'
            ul3_key = f'UL3_APER{suffix_key}'

            # Aperture area
            results[npix_key] = np.full(len(results), aperture.area)

            # When annulus is defined
            if annulus_diameter_pixel is not None:
                annulus_pixel = annulus_diameter_pixel[i] /2
                annulus = CircularAnnulus(positions, r_in=radius_pixel, r_out=annulus_pixel)
                bkg_table = aperture_photometry(data, annulus, mask=mask)
                bkg_area_ratio = aperture.area / annulus.area
                annulus_bkg_flux = bkg_table['aperture_sum'] * bkg_area_ratio

                flux_net = phot_table['aperture_sum'] - annulus_bkg_flux
                results[flux_key] = flux_net
                results[annul_key] = annulus_bkg_flux
                results[mag_key] = -2.5 * np.log10(flux_net)
                results[magannul_key] = -2.5 * np.log10(annulus_bkg_flux)
            # When only aperutre is defined
            else:
                results[flux_key] = phot_table['aperture_sum']
                results[mag_key] = -2.5 * np.log10(phot_table['aperture_sum'])

            # When error is defined
            if 'aperture_sum_err' in phot_table.colnames:
                results[fluxerr_key] = phot_table['aperture_sum_err']
                results[magerr_key] = 2.5 / np.log(10) * phot_table['aperture_sum_err'] / phot_table['aperture_sum']
                
            # Calculation for threshold (when error is defined)
            if bkgrms is not None:
                var_tbl = aperture_photometry(bkgrms**2, aperture, mask=mask)
                results[skysig_key] = np.sqrt(var_tbl['aperture_sum'])
                ul5_flux = 5 * results[skysig_key]
                ul3_flux = 3 * results[skysig_key]
                results[ul5_key] = -2.5 * np.log10(ul5_flux)
                results[ul3_key] = -2.5 * np.log10(ul3_flux)

        cat_path = target_img.savepath.catalogpath.with_suffix('.circ.cat')
        target_catalog = Catalog(path = cat_path, catalog_type = 'all', load = False) 
        target_catalog.data = results
        
        if save:
            target_catalog.write(verbose = verbose)
        else:
            target_catalog.remove(verbose = verbose)
        
        if visualize:
            save_path = None
            if save_fig:
                save_path = str(target_catalog.path) + '.png'
            self._visualize_objects(target_img=target_img, 
                                    objects=target_catalog.data, 
                                    size=100, 
                                    save_path = save_path,
                                    show = visualize)  
        
        if target_bkg is not None:
            target_img_sub.remove(verbose = verbose)
                  
        return target_catalog

    def elliptical_photometry(self,
                              target_img: Union[ScienceImage, ReferenceImage],
                              x_arr: Union[float, list, np.ndarray],  # pixel or RA
                              y_arr: Union[float, list, np.ndarray],  # pixel or Dec
                              sma_arr: Union[float, list, np.ndarray],  # semi-major (arcsec or pixel)
                              smi_arr: Union[float, list, np.ndarray],  # semi-minor (arcsec or pixel)
                              theta_arr: Union[float, list, np.ndarray],  # degrees
                              unit: str = 'pixel',             # 'pixel' or 'coord'
                              annulus_ratio: float = None, 
                              target_bkg: Optional[Background] = None,
                              target_mask: Union[str, Path, np.ndarray] = None,
                              target_bkgrms: Optional[Errormap] = None,
                              # Other parameters
                              save: bool = True,
                              verbose: bool = True,
                              visualize: bool = True,
                              save_fig: bool = False,
                              **kwargs
                              ):
        """
        Perform elliptical aperture photometry.

        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage, CalibrationImage]
            The target image to perform aperture photometry on.
        x_arr : Union[float, list, np.ndarray]
            The x-coordinates of the sources.
        y_arr : Union[float, list, np.ndarray]
            The y-coordinates of the sources.
        sma_arr : Union[float, list, np.ndarray]
            The semi-major axis of the sources.
        smi_arr : Union[float, list, np.ndarray]
            The semi-minor axis of the sources.
        theta_arr : Union[float, list, np.ndarray]
            The position angle of the sources.
        unit : str, optional
            The unit of the coordinates. Defaults to 'pixel'.
        annulus_ratio : float, optional
            The ratio of the annulus diameter to the semi-major axis. Defaults to None.
        target_bkg : Optional[Background], optional
            The background to subtract from the target image. Defaults to None.
        target_mask : Optional[Mask], optional
            The mask to use for the aperture photometry. Defaults to None.
        target_bkgrms : Optional[Errormap], optional
            The background RMS to use for the aperture photometry. Defaults to None.
        save : bool, optional
            Whether to save the aperture photometry catalog. Defaults to True.
        visualize : bool, optional
            Whether to visualize the aperture photometry. Defaults to True.
        save_fig : bool, optional
            Whether to save the aperture photometry figure. Defaults to False.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        target_catalog : Catalog
            The aperture photometry catalog.
        """

        # Step 1: Background subtraction
        if target_bkg is not None:
            target_img = self.background.subtract_background(
                target_img=target_img,
                target_bkg=target_bkg,
                save = False,
                overwrite=False,
                visualize = visualize,
                save_fig = save_fig,
            )

        # Step 2: Prepare image data
        data = self.helper.to_native(target_img.data)
        mask = self.helper.to_native(target_mask.data) if target_mask is not None else None
        bkgrms = self.helper.to_native(target_bkgrms.data) if target_bkgrms is not None else None

        error = None
        if bkgrms is not None:
            error = calc_total_error(data=data, bkg_error=bkgrms, effective_gain=target_img.egain)

        # Step 3: Convert inputs
        x_raw = np.atleast_1d(x_arr)
        y_raw = np.atleast_1d(y_arr)
        sma_raw = np.atleast_1d(sma_arr)
        smi_raw = np.atleast_1d(smi_arr)
        theta = np.radians(np.atleast_1d(theta_arr))  # convert to radians

        # Step 4: Convert (RA, Dec) to (x, y) if needed
        skycoord = None
        if unit == 'coord':
            skycoord = SkyCoord(ra=x_raw, dec=y_raw, unit='deg')
            x, y = skycoord_to_pixel(skycoord, target_img.wcs)
        elif unit == 'pixel':
            wcs = target_img.wcs
            if wcs:
                skycoord = pixel_to_skycoord(x_raw, y_raw, wcs)
            x, y = x_raw, y_raw
        else:
            raise ValueError("coord_type must be either 'pixel' or 'sky'")

        # Step 5: Convert a/b from arcsec to pixels if needed
        pixelscale = np.mean(target_img.pixelscale)  # arcsec/pixel
        if unit == 'coord':
            sma_image = sma_raw / pixelscale
            smi_image = smi_raw / pixelscale
        elif unit == 'pixel':
            sma_image = sma_raw
            smi_image = smi_raw
        else:
            raise ValueError("unit must be either 'pixel' or 'coord'")

        # Step 6: Initialize results table
        results = Table()
        results['X_IMAGE'] = x
        results['Y_IMAGE'] = y
        results['SMA_IMAGE'] = sma_image
        results['SMI_IMAGE'] = smi_image
        results['THETA_IMAGE'] = np.degrees(theta)
        if skycoord:
            results['X_WORLD'] = skycoord.ra.value
            results['Y_WORLD'] = skycoord.dec.value
            results['SMA_WORLD'] = sma_image * pixelscale
            results['SMI_WORLD'] = smi_image * pixelscale

        # Step 8: Aperture photometry
        fluxes, fluxerrs, areas = [], [], []
        apertures = []
        for xi, yi, smai, smii, thetai in zip(x, y, sma_image, smi_image, theta):
            aperture = f ((xi, yi), a=smai, b=smii, theta=thetai)
            apertures.append(aperture)
            tbl = aperture_photometry(data, aperture, error=error, mask=mask)
            fluxes.append(tbl['aperture_sum'][0])
            areas.append(aperture.area)
            if 'aperture_sum_err' in tbl.colnames:
                fluxerrs.append(tbl['aperture_sum_err'][0])
        
        bkgrms_all = []
        if bkgrms is not None:
            for aperture in apertures:
                rms_tbl = aperture_photometry(bkgrms, aperture, mask=mask)
                bkgrms_all.append(rms_tbl['aperture_sum'][0] / aperture.area)
            results['SKYSIG'] = bkgrms_all
            
        results['FLUX_ELIP'] = fluxes
        results['MAG_ELIP'] = -2.5 * np.log10(fluxes)
        results['NPIX_ELIP'] = areas 

        if fluxerrs:
            results['FLUXERR_ELIP'] = fluxerrs
            results['MAGERR_ELIP'] = 2.5 / np.log(10) * np.array(fluxerrs) / np.array(fluxes)

        # Step 9: Annulus subtraction
        if annulus_ratio is not None:
            ann_fluxes, ann_areas = [], []
            for xi, yi, smai, smii, thetai in zip(x, y, sma_image, smi_image, theta):
                annulus = EllipticalAnnulus((xi, yi), a_in=smai, a_out=smai * annulus_ratio,
                                            b_in=smii, b_out=smii * annulus_ratio, theta=thetai)
                tbl = aperture_photometry(data, annulus, error=error, mask=mask)
                ann_fluxes.append(tbl['aperture_sum'][0])
                ann_areas.append(annulus.area)

            bkg_area_ratio = np.array(areas) / np.array(ann_areas)
            annulus_bkg_flux = np.array(ann_fluxes) * bkg_area_ratio
            results['FLUX_ELIP'] = np.array(fluxes) - annulus_bkg_flux
            results['FLUX_EANNULUS'] = annulus_bkg_flux
            results['MAG_ELIP'] = -2.5 * np.log10(results['FLUX_ELIP'])
            results['MAG_EANNULUS'] = -2.5 * np.log10(results['FLUX_EANNULUS'])
        
        cat_path = target_img.savepath.catalogpath.with_suffix('.ellip.cat')
        target_catalog = Catalog(path = cat_path, catalog_type = 'all', load = False) 
        target_catalog.data = results
        
        if save:
            target_catalog.write(verbose = verbose)
        else:
            target_catalog.remove(verbose = verbose)
        
        if visualize:
            save_path = None
            if save_fig:
                save_path = str(target_catalog.path) + '.png'
            self._visualize_objects(target_img=target_img, 
                                    objects=target_catalog.data, 
                                    size=1000, 
                                    save_path = save_path,
                                    show = visualize)

        return target_catalog

    def _visualize_objects(self, 
                           target_img: Union[ScienceImage, ReferenceImage],
                           objects: Table,
                           size: int = 1000,
                           save_path: str = None,
                           show: bool = False):

        data = target_img.data
        h, w = data.shape

        # Step 1: Compute mean position of all sources
        mean_x = np.mean(objects['X_IMAGE'])
        mean_y = np.mean(objects['Y_IMAGE'])

        # Step 2: Find the object closest to that mean position
        dx = np.array(objects['X_IMAGE']) - mean_x
        dy = np.array(objects['Y_IMAGE']) - mean_y
        dist2 = dx**2 + dy**2
        closest_idx = np.argmin(dist2)

        # Use this object to center the zoom-in box
        center_x = float(objects[closest_idx]['X_IMAGE'])
        center_y = float(objects[closest_idx]['Y_IMAGE'])

        # Step 3: Define cropped region centered on that object
        half_box = size // 2
        x_min = int(max(0, center_x - half_box))
        x_max = int(min(w, center_x + half_box))
        y_min = int(max(0, center_y - half_box))
        y_max = int(min(h, center_y + half_box))

        cropped_data = data[y_min:y_max, x_min:x_max]

        # Step 4: Plot setup
        fig, axes = plt.subplots(1, 2, figsize=(16, 8))

        # --- Full image view ---
        m, s = np.mean(data), np.std(data)
        im0 = axes[0].imshow(data, interpolation='nearest', cmap='gray',
                            vmin=m - s, vmax=m + s, origin='lower')
        axes[0].set_title("Full Background-Subtracted Image")
        plt.colorbar(im0, ax=axes[0], fraction=0.03, pad=0.04)

        # Draw red rectangle showing zoomed region
        zoom_box = Rectangle((x_min, y_min), x_max - x_min, y_max - y_min,
                            linewidth=2, edgecolor='red', facecolor='none')
        axes[0].add_patch(zoom_box)

        # --- Zoomed-in region ---
        m_crop, s_crop = np.mean(cropped_data), np.std(cropped_data)
        im1 = axes[1].imshow(cropped_data, interpolation='nearest', cmap='gray',
                            vmin=m_crop - s_crop, vmax=m_crop + s_crop, origin='lower')
        axes[1].set_title(f"Zoomed Region Centered on Closest to Mean Position")

        # Step 5: Draw apertures for all sources in zoomed region
        for obj in objects:
            x, y = float(obj['X_IMAGE']), float(obj['Y_IMAGE'])
            if x_min <= x <= x_max and y_min <= y <= y_max:
                dx_local = x - x_min
                dy_local = y - y_min

                if 'A_IMAGE' in obj.colnames and 'B_IMAGE' in obj.colnames:
                    a = float(obj['A_IMAGE'])
                    b = float(obj['B_IMAGE'])
                    theta = float(obj['THETA_IMAGE']) if 'THETA_IMAGE' in obj.colnames else 0.0
                    patch = Ellipse((dx_local, dy_local), width=6*a, height=6*b, angle=theta,
                                    edgecolor='lime', facecolor='none', linewidth=1.5, alpha=0.6)
                elif 'SMA_IMAGE' in obj.colnames and 'SMI_IMAGE' in obj.colnames:
                    a = float(obj['SMA_IMAGE'])
                    b = float(obj['SMI_IMAGE'])
                    theta = float(obj['THETA_IMAGE']) if 'THETA_IMAGE' in obj.colnames else 0.0
                    patch = Ellipse((dx_local, dy_local), width=6*a, height=6*b, angle=theta,
                                    edgecolor='lime', facecolor='none', linewidth=1.5, alpha=0.6)
                else:
                    patch = Circle((dx_local, dy_local), radius= 5 / target_img.pixelscale[0],
                                edgecolor='lime', facecolor='none', linewidth=1.5, alpha=0.6)

                axes[1].add_patch(patch)

        plt.tight_layout()
        
        
        if save_path is not None:
            plt.savefig(save_path, bbox_inches='tight', dpi=300)
            print(f"Figure saved to {save_path}")
        
        if show:
            plt.show()
            
        plt.close()
    


# Import functions from downloader.py
from .downloader import getimages, download_images_for_pointings

# Import functions from ps1_reference_image.py
# Assuming ps1_reference_image.py contains image processing functions
from .util import generate_pointings

# Import functions from stitching.py
from .stitching import run_swarp, write_images_to_swarp

# Import utility functions from util.py
from .util import degrees_to_hms_dms, plot_rectangle_on_sky, plot_pointings_on_sky

# __all__ = ['getimages', 'generate_pointings', 'run_swarp', 'degrees_to_hms_dms', 'plot_rectangle_on_sky']




class PlatesolveError(Exception):
    """Custom exception raised when plate solving fails."""

    def __init__(self, message, filepath=None, error_code=None):
        super().__init__(message)
        self.filepath = filepath
        self.error_code = error_code

    def __str__(self):
        parts = [super().__str__()]
        if self.filepath:
            parts.append(f"[File: {self.filepath}]")
        if self.error_code is not None:
            parts.append(f"[Code: {self.error_code}]")
        return ' '.join(parts)
    
class MaskingError(Exception):
    """Custom exception raised when Masking fails."""

    def __init__(self, message, filepath=None, error_code=None):
        super().__init__(message)
        self.filepath = filepath
        self.error_code = error_code

    def __str__(self):
        parts = [super().__str__()]
        if self.filepath:
            parts.append(f"[File: {self.filepath}]")
        if self.error_code is not None:
            parts.append(f"[Code: {self.error_code}]")
        return ' '.join(parts)



#%%
from tqdm import tqdm
import inspect
import os
import glob
import subprocess
from multiprocessing import Pool

from ezphot.helper import Helper
#%%

class SDTDataQuerier:
    """
    SDTDataQuerier is a class that syncs data from the source directory to the destination directory.
    
    This class provides
    
    1. Syncing observational data
    
    2. Syncing calibrated data
    
    3. Showing the list of files in the source and destination directories
    
    4. Showing the list of folders in the source and destination directories
    """
    def __init__(self, 
                 ccd : str = 'C361K'):
        self.helper = Helper()
        self.folders = []
        self.ccd = ccd

    def __repr__(self):
        return f"SDTDataQuerier(ccd={self.ccd})\n For help, use 'help(self)' or `self.help()`."

    def help(self):
        # Get all public methods from the class, excluding `help`
        methods = [
            (name, obj)
            for name, obj in inspect.getmembers(self.__class__, inspect.isfunction)
            if not name.startswith("_") and name != "help"
        ]

        # Build plain text list with parameters
        lines = []
        for name, func in methods:
            sig = inspect.signature(func)
            params = [str(p) for p in sig.parameters.values() if p.name != "self"]
            sig_str = f"({', '.join(params)})" if params else "()"
            lines.append(f"- {name}{sig_str}")

        # Final plain text output
        help_text = ""
        print(f"Help for {self.__class__.__name__}\n{help_text}\n\nPublic methods:\n" + "\n".join(lines))
        
    def sync_obsdata(self, 
                     foldername: str,
                     file_pattern: str = '*.fits',
                     ignore_exists: bool = True):
        """
        Syncs all FITS files from a given foldername in the source directory to the destination directory.
        
        This function opens multiple rsync processes to sync the data. (n_processes = len(telescope_ids))
        The destination directory is defined in the helper.config['SDTDATA_OBSDESTDIR']
        
        Parameters
        ----------
        foldername : str
            The foldername to sync.
        file_pattern : str
            The file pattern to sync.
        ignore_exists : bool
            If True, ignore the files that already exist in the destination directory.
        """
        # Step 1: Get telescope directories
        source_pattern = os.path.join(self.helper.config['SDTDATA_OBSSOURCEDIR'],"7DT??", foldername)
        telescope_dirs = glob.glob(source_pattern)
        
        # Step 2: Extract telescope IDs
        telescope_ids = [os.path.basename(os.path.dirname(os.path.normpath(t))) for t in telescope_dirs]

        if not telescope_ids:
            print("No telescope folders found.")

        # Step 3: Move all files to temporary storage in parallel
        with Pool(processes=len(telescope_ids)) as pool:
            dest_folders = pool.starmap(self._run_obsrsync, [(tid, foldername, file_pattern, ignore_exists) for tid in telescope_ids])

    def sync_scidata(self, targetname : str, 
                     file_pattern: str = '*.fits',
                     ignore_exists: bool = True):
        """
        Syncs all FITS files from a given targetname in the source directory to the destination directory.
        
        This function opens multiple rsync processes to sync the data. (n_processes = len(telescope_ids))
        The destination directory is defined in the helper.config['SDTDATA_SCIDESTDIR']
        
        Parameters
        ----------
        targetname : str
            The target name to sync.
        file_pattern : str
            The file pattern to sync.
        ignore_exists : bool
            If True, ignore the files that already exist in the destination directory.
        """
        # Step 1: Get telescope directories
        source_pattern = os.path.join(self.helper.config['SDTDATA_SCISOURCEDIR'], targetname, "7DT??")
        telescope_dirs = glob.glob(source_pattern)
        
        # Step 2: Extract telescope IDs
        telescope_ids = [os.path.basename((os.path.normpath(t))) for t in telescope_dirs]

        if not telescope_ids:
            print("No telescope folders found.")

        # Step 3: Move all files to temporary storage in parallel
        with Pool(processes=len(telescope_ids)) as pool:
            dest_folders = pool.starmap(self._run_scirsync, [(tid, targetname, file_pattern, ignore_exists) for tid in telescope_ids])

    def show_obssourcedata(self, 
                           foldername: str, 
                           show_only_numbers: bool = False,
                           pattern: str = '*.fits'):
        """
        Shows the number or list of FITS files matching a pattern in a given folder across all 7DT?? telescope directories.

        Parameters
        ----------
        foldername : str
            Subfolder name (e.g., filter name) inside each 7DT?? directory.
        show_only_numbers : bool
            If True, return only the number of matched FITS files per telescope.
        pattern : str
            Glob pattern to match FITS files (default: ``'*.fits'``).

        Returns
        -------
        fits_counts : dict
            Dictionary of {telescope_id: count or list of file paths}, sorted by telescope ID.
        """
        import glob
        import os

        fits_counts = {}

        # Find all 7DT?? telescope directories
        telescope_dirs = glob.glob(os.path.join(self.helper.config['SDTDATA_OBSSOURCEDIR'], "7DT??"))
        print(f"Found {len(telescope_dirs)} telescope directories.")

        for telescope_dir in tqdm(telescope_dirs, desc="Searching telescopes..."):
            telescope_id = os.path.basename(telescope_dir)
            folder_path = os.path.join(telescope_dir, foldername)

            if os.path.isdir(folder_path):
                fits_files = glob.glob(os.path.join(folder_path, pattern))

                if show_only_numbers:
                    fits_counts[telescope_id] = len(fits_files)
                else:
                    fits_counts[telescope_id] = sorted(fits_files)

        sorted_fits_counts = {tid: fits_counts[tid] for tid in sorted(fits_counts)}
        total_counts = sum([len(files) if isinstance(files, list) else files for files in sorted_fits_counts.values()])

        if not sorted_fits_counts:
            print("No matched folders found.")
        else:
            print(f"Total files found: {total_counts}")
        return sorted_fits_counts

    def show_obsdestdata(self, 
                         foldername: str, 
                         show_only_numbers: bool = False,
                         pattern: str = '*.fits'):
        """
        Shows the number or list of FITS files matching a pattern in a given folder across all 7DT?? telescopes.

        Parameters
        ----------
        foldername : str
            Subfolder name (e.g., filter name) inside each 7DT?? directory.
        show_only_numbers : bool
            If True, only show counts of matched files.
        pattern : str
            Pattern to match FITS files (e.g., '*.fits', 'calib*.fits').

        Returns
        -------
        dict
            Dictionary keyed by telescope ID, containing counts or file lists.
        """
        import glob
        import os

        fits_counts = {}

        telescope_dirs = glob.glob(os.path.join(self.helper.config['SDTDATA_OBSDESTDIR'], '7DT??', foldername))

        for telescope_dir in telescope_dirs:
            telescope_id = os.path.basename(os.path.dirname(telescope_dir))
            fits_files = glob.glob(os.path.join(telescope_dir, pattern))

            if show_only_numbers:
                fits_counts[telescope_id] = len(fits_files)
            else:
                fits_counts[telescope_id] = sorted(fits_files)

        sorted_fits_counts = {tid: fits_counts[tid] for tid in sorted(fits_counts)}

        if not sorted_fits_counts:
            print("No matched folders found.")
        else:
            print(sorted_fits_counts)

        return sorted_fits_counts


    def show_obssourcefolder(self, 
                             folder_key : str = '*'):
        """
        Shows the contents of the source and destination directories.
        """
        print("Source directory:", os.path.join( self.helper.config['SDTDATA_OBSSOURCEDIR'], "7DT??", folder_key))
        if "*" in folder_key:
            folder_key = folder_key.replace("*", "")

        folders = set()

        for entry in os.scandir( self.helper.config['SDTDATA_OBSSOURCEDIR']):
            if entry.is_dir() and entry.name.startswith("7DT") and len(entry.name) == 5:
                subfolders = {os.path.join(sub.name) for sub in os.scandir(entry.path) if sub.is_dir()}
                folders.update(subfolders)
            
        sorted_folders = sorted(folders)
                
        matched_folders = []
        for folder in sorted_folders:
            if folder_key in folder:
                matched_folders.append(folder)
            else:
                pass
        if not matched_folders:
            print("No matched folders found.")
        else:
            print(f"{len(matched_folders)} folders found.")
            return matched_folders

    def show_obsdestfolder(self, 
                           folder_key : str = None):
        """
        Shows the contents of the source and destination directories.
        """
        print("Source directory:", os.path.join( self.helper.config['SDTDATA_OBSDESTDIR'], "7DT??", folder_key))
        if "*" in folder_key:
            folder_key = folder_key.replace("*", "")

        folders = set()

        for entry in os.scandir( self.helper.config['SDTDATA_OBSDESTDIR']):
            if entry.is_dir() and entry.name.startswith("7DT") and len(entry.name) == 5:
                subfolders = {os.path.join(sub.name) for sub in os.scandir(entry.path) if sub.is_dir()}
                folders.update(subfolders)
                
        sorted_folders = sorted(folders)
            
        matched_folders = []
        for folder in sorted_folders:
            if folder_key in folder:
                matched_folders.append(folder)
            else:
                pass
        if not matched_folders:
            print("No matched folders found.")
        else:
            print(f"{len(matched_folders)} folders found.")
            return matched_folders
        
    def show_scisourcedata(self, 
                           targetname: str, 
                           show_only_numbers: bool = False, 
                           key: str = 'filter',  # 'filter' or 'telescope'
                           file_pattern: str = '*.fits'  # e.g., '*.fits', 'calib*.fits'
                           ):
        """
        Shows the number of FITS files matching a pattern for each specified folder across all 7DT?? directories.

        Parameters
        ----------
        targetname : str
            Target name under SDTDATA_SCISOURCEDIR.
        show_only_numbers : bool
            If True, only show counts instead of filenames.
        key : str
            'filter' (default) to group by filter folders under telescopes, or 'telescope' to group only by telescope.
        file_pattern : str
            File pattern to match FITS files, e.g., '*.fits', 'calib*.fits', '*stack*.fits'.

        Returns
        -------
        dict
            Dictionary of telescope or filter-wise counts or file lists.
        """
        import glob
        import os

        fits_counts = {}

        if key.lower() == 'filter':
            dirs = glob.glob(os.path.join(self.helper.config['SDTDATA_SCISOURCEDIR'], targetname, "7DT??", '*'))
        elif key.lower() == 'telescope':
            dirs = glob.glob(os.path.join(self.helper.config['SDTDATA_SCISOURCEDIR'], targetname, "7DT??"))
        else:
            raise ValueError("Invalid key. Must be 'filter' or 'telescope'.")
        print(f'# of directories: {len(dirs)}')

        for dir in tqdm(dirs, desc="Searching directories..."):
            id_ = os.path.basename(dir)

            if key.lower() == 'filter':
                fits_files = glob.glob(os.path.join(dir, file_pattern))
            else:
                fits_files = glob.glob(os.path.join(dir, "*", file_pattern))

            if id_ not in fits_counts:
                fits_counts[id_] = 0 if show_only_numbers else []

            if show_only_numbers:
                fits_counts[id_] += len(fits_files)
            else:
                fits_counts[id_].extend(sorted(fits_files))

        sorted_fits_counts = {id_: fits_counts[id_] for id_ in sorted(fits_counts)}
        total_counts = sum([len(files) if isinstance(files, list) else files for files in sorted_fits_counts.values()])

        if not sorted_fits_counts:
            print("No matched targets found.")
            return None
        else:
            print(f"Total files found: {total_counts}")
            return sorted_fits_counts
    
    def show_scidestdata(self, 
                         targetname: str, 
                         show_only_numbers: bool = False,
                         key: str = 'filter',  # 'filter' or 'telescope'
                         pattern: str = '*.fits'  # e.g., '*.fits', 'calib*.fits'
                         ):
        """
        Shows the number of FITS files matching a pattern for each specified folder across all 7DT?? directories.

        Parameters
        ----------
        targetname : str
            Target name under SDTDATA_SCIDESTDIR.
        show_only_numbers : bool
            If True, only show counts instead of filenames.
        key : str
            'filter' (default) to group by filter folders under telescopes, or 'telescope' to group only by telescope.
        pattern : str
            File pattern to match FITS files, e.g., '*.fits', 'calib*.fits'.

        Returns
        -------
        dict
            Dictionary of telescope or filter-wise counts or file lists.
        """
        import glob
        import os

        fits_counts = {}

        if key.lower() == 'filter':
            dirs = glob.glob(os.path.join(self.helper.config['SDTDATA_SCIDESTDIR'], targetname, "7DT??", '*'))
        elif key.lower() == 'telescope':
            dirs = glob.glob(os.path.join(self.helper.config['SDTDATA_SCIDESTDIR'], targetname, "7DT??"))
        else:
            raise ValueError("Invalid key. Must be 'filter' or 'telescope'.")

        for dir in dirs:
            id_ = os.path.basename(dir)

            if key.lower() == 'filter':
                fits_files = glob.glob(os.path.join(dir, pattern))
            else:
                fits_files = glob.glob(os.path.join(dir, "*", pattern))

            if id_ not in fits_counts:
                fits_counts[id_] = 0 if show_only_numbers else []

            if show_only_numbers:
                fits_counts[id_] += len(fits_files)
            else:
                fits_counts[id_].extend(sorted(fits_files))

        sorted_fits_counts = {id_: fits_counts[id_] for id_ in sorted(fits_counts)}
        total_counts = sum([len(files) if isinstance(files, list) else files for files in sorted_fits_counts.values()])
        if not sorted_fits_counts:
            print("No matched folders found.")
            return None
        else:
            print(f"Total files found: {total_counts}")
            return sorted_fits_counts

    def show_scisourcefolder(self, 
                             folder_key : str = '*'):
        """
        Shows the contents of the source directory.
        
        Parameters
        ----------
        folder_key : str
            The folder key to show.
            
        Returns
        -------
        list of folder names: list
            List of folder names that match the folder key.
        """
        print("Source directory:", os.path.join( self.helper.config['SDTDATA_SCISOURCEDIR'], folder_key))
        if "*" in folder_key:
            folder_key = folder_key.replace("*", "")

        matched_folders = []
        all_targets = os.listdir(self.helper.config['SDTDATA_SCISOURCEDIR'])
        for target in all_targets:
            if folder_key in target:
                matched_folders.append(target)
        all_matched_folders = set(matched_folders)
        print(f"{len(all_matched_folders)} folders found.")
        return sorted(all_matched_folders)

    def show_obsdestfolder(self, 
                           folder_key : str = '*'):
        """
        Shows the contents of the source and destination directories.
        
        Parameters
        ----------
        folder_key : str
            The folder key to show.
            
        Returns
        -------
        list of folder names: list
            List of folder names that match the folder key.
        """
        print("Source directory:", os.path.join( self.helper.config['SDTDATA_SCIDESTDIR'], folder_key))
        if "*" in folder_key:
            folder_key = folder_key.replace("*", "")

        matched_folders = []
        all_targets = os.listdir(self.helper.config['SDTDATA_SCIDESTDIR'])
        for target in all_targets:
            if folder_key in target:
                matched_folders.append(target)
        all_matched_folders = set(matched_folders)
        return sorted(all_matched_folders)

    def _run_obsrsync(self, telescope_id, foldername, file_pattern='*com.fits', ignore_exists=True):
        """
        Copy ONLY files matching file_pattern from src to dest (recursively).
        """
        src_folder  = os.path.join(self.helper.config['SDTDATA_OBSSOURCEDIR'], telescope_id, foldername)
        dest_folder = os.path.join(self.helper.config['SDTDATA_OBSDESTDIR'],   telescope_id, foldername)

        if not os.path.exists(src_folder):
            print(f"Source folder does not exist: {src_folder}")
            return

        os.makedirs(dest_folder, exist_ok=True)

        # Normalize to list (support multiple patterns if needed)
        includes = [file_pattern] if isinstance(file_pattern, str) else list(file_pattern or [])

        cmd = [
            "rsync", "-av",
            "--info=progress2",
            "--no-inc-recursive",   # lowers memory on big trees
            "--prune-empty-dirs",   # don't create empty dirs at dest
            "--ignore-existing" if ignore_exists else "",
            "--include", "*/"       # allow directory traversal
        ]
        for pat in includes:
            cmd += ["--include", pat]

        # IMPORTANT: exclude everything else
        cmd += ["--exclude", "*", src_folder + "/", dest_folder + "/"]

        print("Running:", " ".join(cmd))
        subprocess.run(cmd, check=True)
        return dest_folder
    
    
    def _run_scirsync(self, telescope_id, targetname, file_pattern='*com.fits', ignore_exists=True):
        """
        Copy ONLY files matching file_pattern from src to dest (recursively).
        """
        src_folder  = os.path.join(self.helper.config['SDTDATA_SCISOURCEDIR'], targetname, telescope_id)
        dest_folder = os.path.join(self.helper.config['SDTDATA_SCIDESTDIR'],   targetname, telescope_id)

        if not os.path.exists(src_folder):
            print(f"Source folder does not exist: {src_folder}")
            return

        os.makedirs(dest_folder, exist_ok=True)

        # Normalize to list (support multiple patterns if needed)
        includes = [file_pattern] if isinstance(file_pattern, str) else list(file_pattern or [])

        cmd = [
            "rsync", "-av",
            "--info=progress2",
            "--no-inc-recursive",   # lowers memory on big trees
            "--prune-empty-dirs",   # don't create empty dirs at dest
            "--ignore-existing" if ignore_exists else "",
            "--include", "*/"       # allow directory traversal
        ]
        for pat in includes:
            cmd += ["--include", pat]

        # IMPORTANT: exclude everything else
        cmd += ["--exclude", "*", src_folder + "/", dest_folder + "/"]

        print("Running:", " ".join(cmd))
        subprocess.run(cmd, check=True)
        return dest_folder
    

#%%
# Example usage:
if __name__ == "__main__":
    foldername = "2025"  # Add required folder keys
    self = SDTDataQuerier()
    #tbl = self.show_destdata('2025-02-10_gain0')
    tile_id_list = set([
    "T22956"
    ])
    folder_key = '*'
    #self.sync_obsdata(foldername)
    #data = self.show_scisourcedata(targetname)
    import time
    # for targetname in tile_id_list:
    #     print(f"Syncing data for target: {targetname}")
    #     self.sync_scidata(targetname = targetname)
    #     time.sleep(10)
    #sync_manager.sync_all_folders(folder_keys)
    #a = self.show_obssourcefolder(folder_key)
    #a = self.show_scisourcefolder(folder_key = '*')


#%%
# Example usage:
if __name__ == "__main__":
    foldername = "2025"  # Add required folder keys
    self = SDTDataQuerier()
    #tbl = self.show_destdata('2025-02-10_gain0')
    tile_id_list = set([
    "T22956"
    ])
    folder_key = '*'
    #self.sync_obsdata(foldername)
    #data = self.show_scisourcedata(targetname)
    import time
    # for targetname in tile_id_list:
    #     print(f"Syncing data for target: {targetname}")
    #     self.sync_scidata(targetname = targetname)
    #     time.sleep(10)
    #sync_manager.sync_all_folders(folder_keys)
    #a = self.show_obssourcefolder(folder_key)
    #a = self.show_scisourcefolder(folder_key = '*')



import os

def write_images_to_swarp(list_file_path, image_files):
    # Create the file list for SWarp input
    # list_file_path = f"{os.path.dirname(output_file)}/images_to_combine.txt"
    with open(list_file_path, 'w') as f:
        for img in image_files:
            f.write(f"{img}\n")
    print(f"{list_file_path} generated")

def run_swarp(list_file_path, path_outim, path_swarp_conf, center_ra_hms, center_dec_dms, xsize=10200, ysize=6800, nthreads=8):
    """Run SWarp to stitch images together."""
    # Example: Construct the system call to swarp
    path_weightim = os.path.splitext(path_outim)[0] + '.weight.fits'

    swarp_command = (
        f'swarp @{list_file_path} -c {path_swarp_conf} '
        f'-IMAGEOUT_NAME {path_outim} '
        f'-WEIGHTOUT_NAME {path_weightim} '
        f'-CENTER "{center_ra_hms},{center_dec_dms}" '
        f'-IMAGE_SIZE {xsize},{ysize} '  # 7DT image size
        f'-NTHREADS {nthreads} '
    )
    print(swarp_command)
    print(f"="*60)
    os.system(swarp_command)


#%%
from pathlib import Path
from typing import List, Tuple, Optional
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from astropy.io import ascii
import astropy
from shapely.geometry import box

from ezphot.skycatalog import SkyCatalog
from ezphot.helper import Helper

#%%

class SkyCatalogUtility:
    """
    Utility class for loading, filtering, and selecting sources from sky catalogs.    
    """
    
    def __init__(self,
                 catalog_archive_path: Optional[str] = None):
        """
        Parameters
        ----------
        catalog_archive_path : Optional[str], optional
            Path to the catalog archive summary file. If not provided, the path will be set to the catalog_dir in the config.
        """
        self.helper = Helper()
        self.catalog_archive_path = Path(catalog_archive_path) if catalog_archive_path else Path(self.helper.config['CATALOG_DIR']) / 'catalog_summary.ascii_fixed_width'

    def get_catalogs(self,
                     ra: float,
                     dec: float,
                     objname: str = None,
                     fov_ra: float = 1.0,
                     fov_dec: float = 1.0,
                     catalog_type: str = 'GAIAXP',
                     fraction_criteria: float = 0.05,
                     query_when_not_archived: bool = True,
                     verbose: bool = False) -> List[SkyCatalog]:
        """
        Load SkyCatalog instances that overlap with a given coordinate and FOV.
        
        Parameters
        ----------
        ra : float
            The right ascension of the target coordinate.
        dec : float
            The declination of the target coordinate.
        objname : str, optional
            The name of the target object.
        fov_ra : float, optional
            The field of view in right ascension.
        fov_dec : float, optional
            The field of view in declination.
        catalog_type : str, optional
            The type of catalog to load.
        fraction_criteria : float, optional
            The fraction of the catalog that must overlap with the target coordinate and FOV.
        query_when_not_archived : bool, optional
            Whether to query the catalog when it is not archived.
        verbose : bool, optional
            Whether to print verbose output.

        Returns
        -------
        catalogs : List[SkyCatalog]
            The list of SkyCatalog instances that overlap with the target coordinate and FOV.
        """
        def make_sky_rectangle(ra_center, dec_center, fov_ra, fov_dec):
            ra_offset = (fov_ra / 2) / np.cos(np.radians(dec_center))
            dec_offset = fov_dec / 2
            return box(ra_center - ra_offset, dec_center - dec_offset,
                       ra_center + ra_offset, dec_center + dec_offset)

        # Load catalog summary
        catalog_path = self.catalog_archive_path
        if catalog_path is None:
            catalog_path = Path(__file__).parent / 'catalog_archive' / 'catalog_summary.ascii_fixed_width'
        summary = ascii.read(catalog_path, format='fixed_width')
        summary = summary[summary['cat_type'] == catalog_type]

        # Fast angular pre-filter
        ra_arr = np.asarray(summary['ra'])
        dec_arr = np.asarray(summary['dec'])
        cos_dec0, sin_dec0 = np.cos(np.radians(dec)), np.sin(np.radians(dec))
        cos_dec, sin_dec = np.cos(np.radians(dec_arr)), np.sin(np.radians(dec_arr))
        delta_ra = np.radians(ra_arr - ra)
        angular_sep = np.degrees(np.arccos(np.clip(sin_dec0 * sin_dec + cos_dec0 * cos_dec * np.cos(delta_ra), -1, 1)))

        search_radius = min(5, 10 * np.sqrt(fov_ra ** 2 + fov_dec ** 2))
        summary_nearby = summary[angular_sep < search_radius]

        # Target polygon
        target_poly = make_sky_rectangle(ra, dec, fov_ra, fov_dec)

        # Intersection filter
        filtered_rows = []
        for row in summary_nearby:
            tile_poly = make_sky_rectangle(row['ra'], row['dec'], row['fov_ra'], row['fov_dec'])
            if tile_poly.intersects(target_poly):
                intersection_fraction = tile_poly.intersection(target_poly).area / target_poly.area
                self.helper.print(f"Catalog {row['objname']} intersection fraction: {intersection_fraction:.2f}", verbose)
                if intersection_fraction >= fraction_criteria:
                    filtered_rows.append(row)

        # Parallel load
        # with ThreadPoolExecutor(max_workers=6) as executor:
        #     catalogs = list(executor.map(self._load_catalog_worker, filtered_rows))
        catalogs = [self._load_catalog_worker(row) for row in filtered_rows]

        if catalogs:
            self.helper.print(f"Found {len(catalogs)} catalogs matching the region (fraction > {fraction_criteria}).", verbose)
        else:
            self.helper.print("No catalogs found.", verbose)
        
        if query_when_not_archived:
            if len(catalogs) == 0:
                skycatalog = SkyCatalog(objname = objname, ra = ra, dec = dec, fov_ra = fov_ra, fov_dec = fov_dec, catalog_type = catalog_type, overlapped_fraction = fraction_criteria, verbose = verbose)
                catalogs = [skycatalog]

        return catalogs

    def select_reference_sources(self,
                                 catalog: SkyCatalog,
                                 mag_lower: Optional[float] = None,
                                 mag_upper: Optional[float] = None,
                                 **kwargs) -> Tuple['astropy.table.Table', list]:
        """
        Select reference sources from a given SkyCatalog based on predefined cuts.
        
        Parameters
        ----------
        catalog : SkyCatalog
            The SkyCatalog instance to select reference sources from.
        mag_lower : Optional[float], optional
            The lower magnitude bound for filtering reference sources.
        mag_upper : Optional[float], optional
            The upper magnitude bound for filtering reference sources.
        **kwargs : dict
            Additional keyword arguments to filter reference sources.

        Returns
        -------
        ref_sources : astropy.table.Table
            The selected reference sources.
        applied_kwargs : list
            The applied keyword arguments.
        """
        if not catalog.data:
            raise RuntimeError(f'No catalog data found for {catalog.objname}')

        # Define default cutlines
        cutlines = {
            'APASS': dict(e_V_mag=[0.001, 0.2], V_mag=[mag_lower, mag_upper]),
            'GAIA': dict(V_flag=[0, 1], V_mag=[mag_lower, mag_upper]),
            'GAIAXP': {"pmra": [-20, 20], "pmdec": [-20, 20], "bp-rp": [0.0, 1.5], "g_mean": [mag_lower, mag_upper]},
            'PS1': {"gFlags": [0, 10], "g_mag": [mag_lower, mag_upper]},
            'SMSS': {"ngood": [5, 999], "class_star": [0.3, 1.0], "g_mag": [mag_lower, mag_upper]}
        }

        if catalog.catalog_type not in cutlines:
            raise ValueError(f"Invalid catalog type: {catalog.catalog_type}")

        cutline = {**cutlines[catalog.catalog_type], **kwargs}

        ref_sources = catalog.data
        applied_kwargs = []
        for key, value in cutline.items():
            if key in ref_sources.colnames:
                applied_kwargs.append({key: [value]})
                ref_sources = ref_sources[(ref_sources[key] >= value[0]) & (ref_sources[key] <= value[1])]

        return ref_sources, applied_kwargs

    
    @staticmethod
    def _load_catalog_worker(row):
        """Worker to load a SkyCatalog from a summary row."""
        return SkyCatalog(objname=row['objname'],
                          catalog_type=row['cat_type'],
                          fov_ra=row['fov_ra'],
                          fov_dec=row['fov_dec'],
                          verbose = False)
# %%
if __name__ == "__main__":
    # Example usage
    from ezphot.utils import DataBrowser
    from ezphot.imageobjects import *
    from ezphot.methods import *
    import glob
    databrowser = DataBrowser('scidata')
    databrowser.observatory = 'RASA36'
    databrowser.telkey = 'RASA36_KL4040_HIGH_1x1'
    databrowser.objname = 'NGC1566'
    stacked_imgset = databrowser.search(pattern='Calib*60.com.fits', return_type='science')
    stacked_imglist = stacked_imgset.target_images
#%%
    # db = DataBrowser('scidata')
    # db.objname = 'NGC1566'
    # db.observatory = 'RASA36'
    
    # target_imgset = db.search('*.com.fits', 'science')
    # target_imglist = target_imgset.target_images
    target_img = stacked_imglist[0]
    ra = target_img.ra
    dec = target_img.dec
    fov_ra = target_img.fovx
    fov_dec = target_img.fovy
    catalog_type = 'APASS'
    utility = SkyCatalogUtility()
    
    catalogs = utility.get_catalogs(ra, dec, fov_ra, fov_dec, catalog_type=catalog_type, verbose=True)

# %%



#%%
import logging
import re
from pathlib import Path
from typing import Union

import numpy as np
from astropy.io import fits
from astropy.io.fits import Header
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
import astropy.units as u

from ezphot.configuration import Configuration
from ezphot.helper import Helper

#%%

class LazyFileHandler(logging.FileHandler):
    def __init__(self, filename, mode='a', encoding=None, delay=True):
        # `delay=True` avoids creating the file immediately
        super().__init__(filename, mode, encoding, delay=delay)
    
class Logger:
    def __init__(self, logger_name):
        self.path = logger_name
        self._log = self.createlogger(logger_name)

    def log(self):
        return self._log

    def createlogger(self, logger_name, logger_level='INFO'):
        logger = logging.getLogger(logger_name)
        if len(logger.handlers) > 0:
            return logger  # Logger already exists

        logger.setLevel(logger_level)
        formatter = logging.Formatter(
            datefmt='%Y-%m-%d %H:%M:%S',
            fmt='[%(levelname)s] %(asctime)-15s | %(message)s'
        )

        # Stream Handler
        streamHandler = logging.StreamHandler()
        streamHandler.setLevel(logger_level)
        streamHandler.setFormatter(formatter)
        logger.addHandler(streamHandler)

        # Lazy File Handler
        fileHandler = LazyFileHandler(filename=logger_name, delay=True)
        fileHandler.setLevel(logger_level)
        fileHandler.setFormatter(formatter)
        logger.addHandler(fileHandler)

        return logger
    
#%%
class BaseImage(Configuration):
    """Base class for loading, managing, and inspecting FITS images.

    Provides lazy-loading of FITS data and header, WCS access, metadata extraction,
    and visualization support. Used as the parent class for all telescope-acquired images
    such as science, reference, and calibration frames.

    Parameters
    ----------
    path : Union[str, Path]
        Path to the FITS image file.
    telinfo : dict, optional
        Telescope information dictionary. If not provided, it will be estimated automatically.
    """

    def __init__(self, path: Union[Path, str], telinfo : dict = None):
        path = Path(path)

        self.helper = Helper()
        self.path = path
        if telinfo is None:
            try:
                telinfo = self.helper.estimate_telinfo(self.path)
            except:
                raise NotImplementedError("WARNING: Telescope information is not found in the configuration (~/ezphot/config/common/CCD.dat). Please provide telinfo manually.")
            
        self.telinfo = telinfo
        self.telkey = self._get_telkey()
        # Initialize or load status
        super().__init__(telkey = self.telkey)

        self._data = None
        self._header = Header()

    def rename(self, new_name: str):
        """Rename the image file and update the internal path.
        
        Parameters
        ----------
        new_name : str
            New name of the image file.
        
        Returns
        -------
        None
        """
        old_path = self.path
        new_path = self.path.parent / new_name

        # If the target exists, remove it (overwrite)
        if new_path.exists():
            new_path.unlink()  # remove the existing file

        old_path.rename(new_path)
        self.path = new_path
        print(f"Renamed {old_path} to {new_path}")

    def clear(self, clear_data: bool = True, clear_header: bool = False, verbose: bool = True):
        """Clear the image data and/or header from memory.
        
        Parameters
        ----------
        clear_data : bool, optional
            If True, clear the image data. Default is True.
        clear_header : bool, optional
            If True, clear the image header. Default is False.
        """
        if clear_data:
            self._data = None
        if clear_header:
            self._header = Header()
        self.helper.print("Cleared data and/or header from memory.", verbose)

    def update_header(self, **kwargs):
        """Update FITS header values using known key variants.
        
        Parameters
        ----------
        kwargs : dict
            Keyword arguments to update the FITS header.
        
        Returns
        -------
        None
        """
        if self._header is None:
            print("WARNING: Header is not loaded. Cannot update.")
            return
        else:
            for key, value in kwargs.items():
                key_upper = key.upper()
                if key_upper in self._key_variants.keys():
                    key_variants = self._key_variants[key_upper]
                    for key_variant in key_variants:
                        if key_variant in self._header:
                            self._header[key_variant] = value
                else:
                    print(f'WARNING: Key {key} not found in key_variants.')

    def load(self):
        """Load both image data and header from disk.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if not self.is_exists:
            raise FileNotFoundError(f'File not found: {self.path}')
        self.data
        self.header
        
    def show(self, 
            coord_type: str = "pixel",
            cmap='gray', 
            scale='zscale', 
            downsample=4, 
            figsize=(8, 6), 
            title=None, 
            save_path: str = False,
            close_fig: bool = False):
        """
        Visualize the FITS image with optional WCS overlay and downsampling.
        """
        import numpy as np
        import matplotlib.pyplot as plt
        from astropy.visualization import ZScaleInterval, MinMaxInterval
        from astropy import units as u

        # ----------------------------
        # 1. Prepare data and WCS
        # ----------------------------
        if self.data is None:
            print("WARNING: Image data is not loaded. Please load the image first.")
            return

        data = np.nan_to_num(self.data, nan=0.0, posinf=0.0, neginf=0.0)
        w = getattr(self, "wcs", None)

        if downsample > 1:
            data = data[::downsample, ::downsample]
            if coord_type == "coord" and w is not None:
                w = w.deepcopy()
                if w.wcs.has_cd():
                    w.wcs.cd *= downsample
                elif w.wcs.has_pc():
                    w.wcs.cdelt *= downsample
                else:
                    w.wcs.cdelt *= downsample
                w.wcs.crpix = (w.wcs.crpix - 1) / downsample + 1

        # ----------------------------
        # 2. Prepare visualization
        # ----------------------------
        interval = ZScaleInterval() if scale == 'zscale' else MinMaxInterval()
        vmin, vmax = interval.get_limits(data)

        use_wcs = (coord_type == "coord" and w is not None)

        # ----------------------------
        # 3. Plotting
        # ----------------------------
        if use_wcs:
            try:
                fig, ax = plt.subplots(figsize=figsize, subplot_kw={'projection': w})
                ax.grid(color='white', ls='dotted')

                # Force decimal degree ticks with 0.1 degree spacing
                lon, lat = ax.coords[0], ax.coords[1]
                lon.set_format_unit(u.deg)
                lat.set_format_unit(u.deg)
                
                lon.set_major_formatter('d.dd')  # RA with 4 decimals
                lat.set_major_formatter('d.d')    # Dec with 3 decimals
                
                # Set grid spacing to get exactly 5 grid lines for each axis
                # Calculate spacing based on field of view
                fov_ra = self.fovx  # Field of view in RA (degrees)
                fov_dec = self.fovy  # Field of view in Dec (degrees)
                
                # Correct RA spacing for declination (RA lines converge toward poles)
                center = self.center
                if center['dec'] is not None:
                    dec_rad = np.radians(center['dec'])
                    ra_spacing = (fov_ra / 4) / np.cos(dec_rad)
                else:
                    ra_spacing = fov_ra / 4
                
                dec_spacing = fov_dec / 4
                
                lon.set_ticks(spacing=ra_spacing * u.deg)
                lat.set_ticks(spacing=dec_spacing * u.deg)
                
                # Rotate x-axis tick labels by 45 degrees
                ax.tick_params(axis='x', rotation=45)

                ax.set_xlabel("RA [deg]")
                ax.set_ylabel("Dec [deg]")
            except Exception as e:
                print(f"WARNING: WCS plotting failed ({e}). Falling back to pixel coords.")
                fig, ax = plt.subplots(figsize=figsize)
                use_wcs = False
        else:
            fig, ax = plt.subplots(figsize=figsize)

        # Show image
        img = ax.imshow(data, cmap=cmap, origin='lower', vmin=vmin, vmax=vmax)

        # Colorbar
        cbar = fig.colorbar(img, ax=ax, orientation='vertical', fraction=0.046, pad=0.04)
        cbar.set_label("Pixel value")

        # Pixel coordinate ticks
        if not use_wcs:
            yticks = np.linspace(0, data.shape[0]-1, num=6, dtype=int)
            xticks = np.linspace(0, data.shape[1]-1, num=6, dtype=int)
            ax.set_yticks(yticks)
            ax.set_yticklabels((yticks * downsample).astype(int))
            ax.set_xticks(xticks)
            ax.set_xticklabels((xticks * downsample).astype(int))
            ax.set_xlabel("X pixel")
            ax.set_ylabel("Y pixel")

        # Title
        ax.set_title(title or self.path.name)

        plt.tight_layout()
        if save_path:
            plt.savefig(save_path, dpi=300)
            print(f"Saved: {save_path}")
        if close_fig:
            plt.close(fig)

        return fig, ax

    def show_position(self,
                    x: float,
                    y: float,
                    radius_arcsec: float = None,
                    a_arcsec: float = None,
                    b_arcsec: float = None,
                    theta_deg: float = None,
                    coord_type: str = 'pixel',
                    downsample: int = 1,
                    zoom_radius_pixel: int = 100,
                    cmap: str = 'gray',
                    scale: str = 'zscale',
                    figsize=(6, 6),
                    ax=None,
                    save_path: str = None,
                    title: bool = True):
        """
        Display a zoomed-in view around a given (x, y) or (RA, Dec) position.

        Parameters
        ----------
        x : float
            X-coordinate (pixel or RA)
        y : float
            Y-coordinate (pixel or Dec)
        radius_arcsec : float, optional
            Radius of the circle in arcseconds. Default is None.
        a_arcsec : float, optional
            Semi-major axis of the ellipse in arcseconds. Default is None.
        b_arcsec : float, optional
            Semi-minor axis of the ellipse in arcseconds. Default is None.
        theta_deg : float, optional
            Position angle of the ellipse in degrees. Default is None.
        coord_type : str, optional
            'pixel' or 'coord'. Default is 'pixel'.
        downsample : int, optional
            Step size for downsampling via slicing. Default is 4.
        zoom_radius_pixel : int, optional
            Radius of the zoom-in region in pixels. Default is 100.
        cmap : str, optional
            Matplotlib colormap. Default is 'gray'.
        scale : str, optional
            Scaling method. Default is 'zscale'.
        figsize : tuple, optional
            Matplotlib figure size. Default is (6, 6).
        ax : matplotlib Axes object, optional
            If provided, the image will be plotted on this axis. Default is None.
        save_path : str, optional
            Path to save the figure. Default is None.
        title : bool, optional
            Whether to display the title. Default is True.
        """
        import matplotlib.pyplot as plt
        from astropy.visualization import ZScaleInterval, MinMaxInterval
        from matplotlib.patches import Circle

        data = self.data
        wcs = self.wcs
        if data is None:
            print("No image data loaded.")
            return

        # Convert (RA, Dec) to pixel if needed
        if coord_type == 'coord':
            if wcs is None:
                print("No valid WCS for sky-to-pixel conversion.")
                return
            from astropy.coordinates import SkyCoord
            import astropy.units as u
            coord = SkyCoord(x * u.deg, y * u.deg, frame='icrs')
            x_pix, y_pix = wcs.world_to_pixel(coord)
        elif coord_type == 'pixel':
            x_pix, y_pix = x, y
        else:
            raise ValueError("coord_type must be 'pixel' or 'coord'.")

        x_pix, y_pix = int(x_pix), int(y_pix)

        # Extract zoom window
        x_min = max(0, x_pix - zoom_radius_pixel)
        x_max = min(data.shape[1], x_pix + zoom_radius_pixel)
        y_min = max(0, y_pix - zoom_radius_pixel)
        y_max = min(data.shape[0], y_pix + zoom_radius_pixel)
        size = x_max - x_min
        if radius_arcsec is None:
            radius = size * (0.08/downsample)
        else:
            radius = radius_arcsec / np.mean(self.pixelscale)
        cutout = data[y_min:y_max:downsample, x_min:x_max:downsample]

        # Scaling
        interval = ZScaleInterval() if scale == 'zscale' else MinMaxInterval()
        vmin, vmax = interval.get_limits(cutout)

        # Draw
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = ax.figure

        ax.imshow(cutout, cmap=cmap, origin='lower', vmin=vmin, vmax=vmax)
        if a_arcsec is not None and b_arcsec is not None and theta_deg is not None:
            from matplotlib.patches import Ellipse
            ax.add_patch(Ellipse(
                ((x_pix - x_min) // downsample, (y_pix - y_min) // downsample),
                width=6*a_arcsec/np.mean(self.pixelscale), height=6*b_arcsec/np.mean(self.pixelscale), angle=theta_deg,
                edgecolor='red', facecolor='none', linewidth=0.5
            ))
        else:
            ax.add_patch(Circle(
                ((x_pix - x_min) // downsample, (y_pix - y_min) // downsample),
                radius=radius, edgecolor='red', facecolor='none', linewidth=0.5
            ))
        ax.axis('off')
        ax.set_aspect('auto')  # ? Avoid square enforcement

        if title is not None:
            ax.set_title(f"{title}", fontsize=8, pad=1)  # Less padding

        if save_path:
            fig.savefig(save_path, dpi=300)
            print(f"Saved: {save_path}")

        return fig, ax

    def run_ds9(self):
        """Open the image with SAOImage DS9.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.helper.run_ds9(self.path)        
        
    @property
    def data(self):
        """Image data array."""
        if not self.is_data_loaded and self.is_exists:
            try:
                self._data = fits.getdata(self.path)
            except Exception as e:
                print(f"Failed to load data from {self.path}: {e}")
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def header(self):
        """FITS header object."""
        if not self.is_header_loaded and self.is_exists:
            try:
                self._header = fits.getheader(self.path)
            except Exception as e:
                print(f"Failed to load header from {self.path}: {e}")
        return self._header

    @header.setter
    def header(self, value):
        self._header = value
         
    @property
    def is_data_loaded(self):
        """Whether the image data is loaded."""
        return self._data is not None

    @property
    def is_header_loaded(self):
        """Whether the image header is loaded."""
        return isinstance(self._header, Header) and len(self._header) > 0

    @property
    def is_exists(self):
        """Whether the image file exists."""
        return self.path.exists()
        
    @property
    def observatory(self):
        """ Name of the observatory."""
        return str(self.telinfo['obs'])
    
    @property
    def telname(self):
        """ Name of the telescope."""
        header = self._header
        for key in self._key_variants['TELESCOP']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def imgtype(self):
        """ Type of the image. Among BIAS, DARK, FLAT, LIGHT, OBJECT, UNKNOWN."""
        header = self._header
        for key in self._key_variants['IMGTYPE']:
            if key in header:
                imgtype = header[key]
                imgtype_variants = dict(BIAS= ['BIAS', 'Bias', 'bias', 'ZERO', 'Zero', 'zero'],
                                        DARK= ['DARK', 'Dark', 'dark'],
                                        FLAT= ['FLAT', 'Flat', 'flat'],
                                        LIGHT= ['LIGHT', 'Light', 'light', 'OBJECT', 'Object', 'object'])
                for key, variants in imgtype_variants.items():
                    if imgtype in variants:
                        return key
        print('WARNING: IMGTYPE not found in header')
        return 'UNKNOWN'
    
    @property
    def altitude(self):
        """ Altitude of the telescope.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        altitude : float
            Altitude of the telescope.
        """
        header = self._header
        for key in self._key_variants['ALTITUDE']:
            if key in header:
                return float(header[key])
        return None

    
    @property
    def azimuth(self):
        """Azimuth of the telescope."""
        header = self._header
        for key in self._key_variants['AZIMUTH']:
            if key in header:
                return float(header[key])
        return None

    @property 
    def ra(self):
        """Right ascension of the target."""
        header = self._header
        for key in self._key_variants['RA']:
            if key in header:
                return float(header[key])
        return None
        
    @property
    def dec(self):
        """Declination of the target."""
        header = self._header
        for key in self._key_variants['DEC']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def objname(self):
        """Object name."""
        header = self._header
        for key in self._key_variants['OBJECT']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def obsmode(self):
        """Observation mode."""
        header = self._header
        for key in self._key_variants['OBSMODE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def specmode(self):
        """Spectroscopic mode."""
        header = self._header
        for key in self._key_variants['SPECMODE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def filter(self):
        """Filter name."""
        header = self._header
        for key in self._key_variants['FILTER']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def ccdtemp(self):
        """CCD temperature."""
        header = self._header
        for key in self._key_variants['CCD-TEMP']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def obsdate(self):
        """Observation date in UTC."""
        header = self._header
        # First, search for utcdate
        for key in self._key_variants['DATE-OBS']:
            if key in header:
                return Time(header[key], format = 'isot', scale = 'utc').isot
        # If not found, search for jd
        for key in self._key_variants['JD']:
            if key in header:
                return Time(header[key], format = 'jd').isot
        # If not found, search for mjd
        for key in self._key_variants['MJD']:
            if key in header:
                return Time(header[key], format = 'mjd').isot
        # If not found, estimate from filename
        try:
            name = self.path.stem
            # Try YYYYMMDD or YYYY-MM-DD or YYYY_MM_DD
            m = re.search(r'(?<!\d)(20\d{2})[-_]?([01]\d)[-_]?([0-3]\d)(?!\d)', name)
            if m:
                year = int(m.group(1))
                month = int(m.group(2))
                day = int(m.group(3))
            else:
                # Try YYMMDD or YY-MM-DD or YY_MM_DD
                m = re.search(r'(?<!\d)(\d{2})[-_]?([01]\d)[-_]?([0-3]\d)(?!\d)', name)
                if m:
                    yy = int(m.group(1))
                    year = 2000 + yy
                    month = int(m.group(2))
                    day = int(m.group(3))
                else:
                    return None

            # Basic validation of month/day ranges
            if not (1 <= month <= 12 and 1 <= day <= 31):
                return None

            datestr = f"{year:04d}-{month:02d}-{day:02d}T00:00:00"
            return Time(datestr, format='isot', scale='utc').isot
        except Exception:
            return None
    
    @property
    def mjd(self):
        """Modified Julian date of the observation."""
        return Time(self.obsdate, format='isot').mjd
    
    @property
    def jd(self):
        """Julian date of the observation."""
        return Time(self.obsdate, format='isot').jd

    @property
    def exptime(self):
        """Exposure time of the image."""
        header = self._header
        for key in self._key_variants['EXPTIME']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def binning(self):
        """Binning of the image."""
        header = self._header
        for key in self._key_variants['BINNING']:
            if key in header:
                return int(header[key])
        return None
        
    @property
    def gain(self):
        """Gain of the image."""
        header = self._header
        for key in self._key_variants['GAIN']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def egain(self):
        """Electron gain of the image."""
        header = self._header
        for key in self._key_variants['EGAIN']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def naxis1(self):
        """Number of pixels along the first axis."""
        header = self._header
        for key in self._key_variants['NAXIS1']:
            if key in header:
                return int(header[key])
        return None

    @property
    def naxis2(self):
        """Number of pixels along the second axis."""
        header = self._header
        for key in self._key_variants['NAXIS2']:
            if key in header:
                return int(header[key])
        return None
    
    @property
    def ncombine(self):
        """Number of combined images."""
        header = self._header
        for key in self._key_variants['NCOMBINE']:
            if key in header:
                return int(header[key])
        return None
    
    @property
    def biaspath(self):
        """Path to the bias image."""
        header = self._header
        for key in self._key_variants['BIASPATH']:
            if key in header:
                return str(header[key])
        return None
            
    @property
    def darkpath(self):
        """Path to the dark image."""
        header = self._header
        for key in self._key_variants['DARKPATH']:
            if key in header:
                return str(header[key])
        return None

    @property
    def flatpath(self):
        """Path to the flat image."""
        header = self._header
        for key in self._key_variants['FLATPATH']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def maskpath(self):
        """Path to the mask image."""
        header = self._header
        for key in self._key_variants['MASKPATH']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def masktype(self):
        """Type of the mask image."""
        header = self._header
        for key in self._key_variants['MASKTYPE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def bkgpath(self):
        """Path to the background image."""
        header = self._header
        for key in self._key_variants['BKGPATH']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def bkgtype(self):
        """Type of the background image."""
        header = self._header
        for key in self._key_variants['BKGTYPE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def emappath(self):
        """Path to the emap image."""
        header = self._header
        for key in self._key_variants['EMAPPATH']:
            if key in header:
                return str(header[key])
        return None

    @property
    def emaptype(self):
        """Type of the emap image."""
        header = self._header
        for key in self._key_variants['EMAPTYPE']:
            if key in header:
                return str(header[key])
        return None
            
    @property
    def fovx(self):
        """Field of view along the first axis."""
        pixscale = self.pixelscale
        if pixscale is not None:
            fovx = pixscale[0] * self.naxis1 / 3600  # Convert to degrees
            return float('%.3f' % fovx)
        else:
            return float('%.3f' %(self.telinfo['pixelscale'] * self.naxis1 / 3600))
    
    @property
    def fovy(self):
        """Field of view along the second axis."""
        pixscale = self.pixelscale
        if pixscale is not None:
            fovy = pixscale[1] * self.naxis2 / 3600  # Convert to degrees
            return float('%.3f' % fovy)
        else:
            return float('%.3f' %(self.telinfo['pixelscale'] * self.naxis2 / 3600))
    
    @property
    def wcs(self):
        """WCS information of the image."""
        try:
            wcs = WCS(self._header)
            if self.imgtype.upper() in ['BIAS', 'DARK', 'FLAT']:
                raise ValueError("WCS is not available for bias, dark, or flat images.")
            return wcs
        except:
            return None
    
    @property
    def center(self):
        """Center pixel (0-based) and its world coordinates (RA, Dec)."""
        x_center = (self.naxis1 - 1) / 2
        y_center = (self.naxis2 - 1) / 2
        ra = dec = None

        if self.wcs is not None:
            try:
                skycoord = self.wcs.pixel_to_world(x_center, y_center)
                ra = skycoord.ra.deg
                dec = skycoord.dec.deg
            except Exception as e:
                print(f"WCS conversion failed: {e}")

        return {'x': x_center, 'y': y_center, 'ra': ra, 'dec': dec}
        
    @property
    def pixelscale(self):
        """Pixel scale of the image."""
        try:
            return proj_plane_pixel_scales(self.wcs) * 3600  # Convert to arcseconds
        except:
            return None
    
    @property
    def zp(self):
        """Zero point of the image."""
        header = self._header
        for key in self._key_variants['ZP']:
            if key in header:
                return float(header[key])
        return None

    @property
    def zperr(self):    
        """Zero point error of the image."""
        header = self._header
        for key in self._key_variants['ZPERR']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def depth(self):
        """Depth of the image."""
        header = self._header
        for key in self._key_variants['DEPTH']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def seeing(self):
        """Seeing of the image."""
        header = self._header
        for key in self._key_variants['SEEING']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def ncombine(self):
        """Number of combined images."""
        header = self._header
        for key in self._key_variants['NCOMBINE']:
            if key in header:
                return int(header[key])
        return None
        
    @property
    def _key_variants(self):
        '''Key variants of the FITS header.'''
        # Define key variants, if a word is duplicated in the same variant, posit the word with the highest priority first
        key_variants_upper = {
            # Observation information
            'ALTITUDE': ['ALT', 'ALTITUDE', 'CENTALT'],
            'AZIMUTH': ['AZ', 'AZIMUTH', 'CENTAZ'],
            'GAIN': ['GAIN'],
            'EGAIN': ['EGAIN'],
            'CCD-TEMP': ['CCDTEMP', 'CCD-TEMP'],
            'FILTER': ['FILTER', 'FILTNAME', 'BAND'],
            'IMGTYPE': ['IMGTYPE', 'IMAGETYP', 'IMGTYP'],
            'EXPTIME': ['EXPTIME', 'EXPOSURE'],
            'DATE-OBS': ['DATE-OBS', 'OBSDATE', 'UTCDATE'],
            'DATE-LOC': ['DATE-LOC', 'DATE-LTC', 'LOCDATE', 'LTCDATE'],
            'JD' : ['JD', 'JD-HELIO', 'JD-UTC', 'JD-OBS'],
            'MJD' : ['MJD', 'MJD-HELIO', 'MJD-UTC', 'MJD-OBS'],
            'RA': ['CRVAL1', 'RA', 'OBJCTRA', 'OBSRA'],
            'DEC': ['CRVAL2', 'DEC', 'DECL', 'DEC.', 'DECL.', 'CRVAL2', 'OBJCTDEC', 'OBSDEC'],   
            'TELESCOP' : ['TELESCOP', 'TELNAME'],
            'BINNING': ['BINNING', 'XBINNING'],
            'OBJECT': ['OBJECT', 'OBJNAME', 'TARGET', 'TARNAME'],
            'OBJCTID': ['OBJCTID', 'OBJID', 'ID'],
            'OBSMODE': ['OBSMODE', 'MODE'],
            'SPECMODE': ['SPECMODE'],
            'NTELESCOP': ['NTELESCOP', 'NTEL'],
            'NCOMBINE': ['NCOMBINE', 'NCOMB'],
            'NOTE': ['NOTE'],
            'NAXIS1': ['NAXIS1'],
            'NAXIS2': ['NAXIS2'],
            # Additional key after processing
            'CTYPE1': ['CTYPE1'],
            'CTYPE2': ['CTYPE2'],
            'CRVAL1': ['CRVAL1'],
            'CRVAL2': ['CRVAL2'],
            'SEEING': ['SEEING'],
            'ELONGATION': ['ELONGATION', 'ELONG'],
            'SKYSIG': ['SKYSIG', 'SKY_SIG'],
            'SKYVAL': ['SKYVAL', 'SKY_VAL'],
            'ZP': ['ZP_AUTO', 'ZP_2'],
            'ZPERR': ['ZPERR_AUTO', 'EZP_2'],
            'DEPTH': ['UL5SKY_APER_1', 'UL5SKY_APER_2', 'UL5_4'],
            # Path information
            'SAVEPATH': ['SAVEPATH'],
            'BIASPATH': ['BIASPATH'],
            'DARKPATH': ['DARKPATH'],
            'FLATPATH': ['FLATPATH'],
            'BKGPATH': ['BKGPATH'],
            'BKGTYPE': ['BKGTYPE'],
            'EMAPPATH': ['EMAPPATH'],
            'EMAPTYPE': ['EMAPTYPE'],
            'MASKPATH': ['MASKPATH'],
            'MASKTYPE': ['MASKTYPE'],
        }

        return key_variants_upper

    def _get_telkey(self):
        """ Get the telescope name from the FITS header """
        telinfo = self.telinfo
        if telinfo['mode']:
            telkey = f"{telinfo['obs']}_{telinfo['ccd']}_{telinfo['mode']}_{telinfo['binning']}x{telinfo['binning']}"
        else:
            telkey = f"{telinfo['obs']}_{telinfo['ccd']}_{telinfo['binning']}x{telinfo['binning']}"
        return telkey

# %%